"""Module for forcing data input and output operations."""
from pathlib import Path
import dask
import hdf5storage
import numpy as np
import xarray as xr
from PyStemmusScope import global_data
from PyStemmusScope import utils
from PyStemmusScope import variable_conversion as vc


def _write_matlab_ascii(fname, data, ncols):
    """Write data in the Matlab ascii format.

    Equivalent to `save([-], '-ascii')` in Matlab.

    Args:
        fname (path or str): Path to the file that should be written
        data (np.array): Array with data to write to file
        ncols (int, optional): The number of data columns, required to correctly format
            the ascii file when writing multiple variables.
    """
    matlab_fmt = " %14.7e"
    multi_fmt = [matlab_fmt] * ncols
    multi_fmt[0] = f" {multi_fmt[0]}"
    np.savetxt(fname, data, multi_fmt)


def read_forcing_data_plumber2(forcing_file: Path, start_time: str, end_time: str):
    """Read the forcing data from the provided netCDF file and apply unit conversion.

    Args:
        forcing_file: Path to the netCDF file containing the forcing data
        start_time: Start of time range in ISO format string, e.g. ,
            'YYYY-MM-DDTHH:MM:SS'.
        end_time: End of time range in ISO format string, e.g.,
            'YYYY-MM-DDTHH:MM:SS'.

    Returns:
        dict: Dictionary containing the different variables required by STEMMUS_SCOPE
            for the different forcing files.
    """
    ds_forcing = xr.open_dataset(forcing_file)

    # check if time range is covered by forcing
    # if so, return a subset of forcing matching the given time range
    ds_forcing = _slice_forcing_file(
        ds_forcing,
        start_time,
        end_time,
    )

    # remove the x and y coordinates from the data variables to make the numpy arrays 1D
    ds_forcing = ds_forcing.squeeze(["x", "y"])

    data = {}

    # Expected time format is days (as floating point) since Jan 1st 00:00.
    data["doy_float"] = (
        ds_forcing["time"].dt.dayofyear
        - 1
        + ds_forcing["time"].dt.hour / 24
        + ds_forcing["time"].dt.minute / 60 / 24
    )
    data["year"] = ds_forcing["time"].dt.year.astype(float)

    data["t_air_celcius"] = ds_forcing["Tair"] - 273.15  # conversion from K to degC.
    data["psurf_hpa"] = ds_forcing["Psurf"] / 100  # conversion from Pa to hPa
    data["co2_conv"] = (
        vc.co2_molar_fraction_to_kg_per_m3(
            ds_forcing["CO2air"] * 1e-6  # ppm -> molar fraction
        )
        * 1e6
    )  # kg/m3 -> mg/m3
    data["precip_conv"] = ds_forcing["Precip"] / 10  # conversion from mm/s to cm/s
    data["lw_down"] = ds_forcing["LWdown"]
    data["sw_down"] = ds_forcing["SWdown"]
    data["wind_speed"] = vc.mask_data(ds_forcing["Wind"], min_value=0.05)
    data["rh"] = ds_forcing["RH"]
    data["vpd"] = ds_forcing["VPD"]
    data["lai"] = vc.mask_data(ds_forcing["LAI"], min_value=0.01)

    # calculate ea, conversion from kPa to hPa:
    data["ea"] = vc.calculate_ea(data["t_air_celcius"], data["rh"]) * 10

    # Load in non-timedependent variables
    data["sitename"] = forcing_file.name.split("_")[0]

    # Forcing data timestep size in seconds
    time_delta = (
        ds_forcing.time.values[1] - ds_forcing.time.values[0]
    ) / np.timedelta64(1, "s")
    data["DELT"] = time_delta.astype(float)
    data["total_timesteps"] = ds_forcing.time.size

    data["latitude"] = ds_forcing["latitude"].values
    data["longitude"] = ds_forcing["longitude"].values
    data["elevation"] = ds_forcing["elevation"].values
    data["IGBP_veg_long"] = np.repeat(
        ds_forcing["IGBP_veg_long"].values, ds_forcing.time.size
    ).T
    data["reference_height"] = ds_forcing["reference_height"].values
    data["canopy_height"] = ds_forcing["canopy_height"].values

    # these are needed by save.py
    data["time"] = ds_forcing["time"]
    data["Qair"] = ds_forcing["Qair"]

    return data


def read_forcing_data_global(  # noqa:PLR0913 (too many arguments)
    global_data_dir: Path,
    lat: float,
    lon: float,
    start_time: np.datetime64,
    end_time: np.datetime64,
    timestep: str = "1800s",
) -> dict:
    """Read forcing data for a certain location, based on global datasets.

    Args:
        global_data_dir: Path to the directory containing the global datasets.
        lat: Latitude of the site of interest.
        lon: Longitude of the site of interest.
        start_time: Start time of the model run.
        end_time: End time of the model run.
        timestep: Desired timestep of the model, this is derived from the forcing data.
            In a pandas-timedelta compatible format. Defaults to "1800S" (half an hour).

    Returns:
        Dictionary containing the forcing data.
    """
    # see https://docs.dask.org/en/latest/array-slicing.html#efficiency
    with dask.config.set(**{"array.slicing.split_large_chunks": True}):  # type: ignore
        return global_data.collect_datasets(
            global_data_dir=global_data_dir,
            latlon=(lat, lon),
            time_range=(start_time, end_time),
            timestep=timestep,
        )


def write_dat_files(data: dict, input_dir: Path):
    """Fuction to write the single-data .dat files for the STEMMUS_SCOPE matlab model.

    Args:
        data: Data dictionary generated by read_forcing
        input_dir: Directory to which the different single-column .dat files should be
            written to.
    """
    write_info = {
        "doy_float": "t_.dat",
        "t_air_celcius": "Ta_.dat",
        "sw_down": "Rin_.dat",
        "lw_down": "Rli_.dat",
        "psurf_hpa": "p_.dat",
        "wind_speed": "u_.dat",
        "co2_conv": "CO2_.dat",
        "ea": "ea_.dat",
        "year": "year_.dat",
    }
    for var, fname in write_info.items():
        _write_matlab_ascii(input_dir / fname, data[var], ncols=1)


def write_lai_file(data: dict, fpath: Path):
    """Write the ascii LAI_.dat file for STEMMUS_SCOPE.

    Args:
        data: Dictionary containing the required variables. Generated by the
            function `read_forcing_data`.
        fpath: Full path, including filename, to which the file should be
            written to.
    """
    lai_file_data = np.vstack([data["doy_float"], data["lai"]]).T
    _write_matlab_ascii(fpath, lai_file_data, ncols=2)


def write_meteo_file(data: dict, fpath: Path):
    """Write the ascii Mdata.txt meteo file for STEMMUS_SCOPE.

    Args:
        data: Dictionary containing the required variables. Generated by the
            function `read_forcing_data`.
        fpath: Full path, including filename, to which the file should be
            written to.
    """
    meteo_data_vars = [
        "doy_float",
        "t_air_celcius",
        "rh",
        "wind_speed",
        "psurf_hpa",
        "precip_conv",
        "sw_down",
        "lw_down",
        "vpd",
        "lai",
    ]
    meteo_file_data = np.vstack([data[var] for var in meteo_data_vars]).T
    _write_matlab_ascii(fpath, meteo_file_data, ncols=len(meteo_data_vars))


def prepare_global_variables(data: dict, input_path: Path):
    """Read and calculate global variables for STEMMUS_SCOPE from forcing data.

    Data will be written to a Matlab binary file (v7.3), under the name
    'forcing_globals.mat' in the specified input directory.

    Args:
        data: Dictionary containing the required variables. Generated by the
            function `read_forcing_data`.
        input_path: Path to which the file should be written to.
    """
    matfile_vars = [
        "latitude",
        "longitude",
        "elevation",
        "IGBP_veg_long",
        "reference_height",
        "canopy_height",
        "DELT",
        "sitename",
    ]
    matfiledata = {key: data[key] for key in matfile_vars}

    matfiledata["Dur_tot"] = float(data["total_timesteps"])  # Matlab expects a 'double'

    hdf5storage.savemat(
        input_path / "forcing_globals.mat", matfiledata, appendmat=False
    )
    utils.remove_dates_from_header(input_path / "forcing_globals.mat")


def prepare_forcing(config: dict) -> None:
    """Prepare the forcing files required by STEMMUS_SCOPE.

    The input directory should be taken from the model configuration file.
    A subset of forcing file will be generated if the time range is covered
    by the time of existing forcing file.

    Args:
        config (dict): The PyStemmusScope configuration dictionary.
    """
    input_path = Path(config["InputPath"])

    loc, fmt = utils.check_location_fmt(config["Location"])

    if fmt == "site":
        forcing_file = utils.get_forcing_file(config)
        data = read_forcing_data_plumber2(
            forcing_file=forcing_file,
            start_time=config["StartTime"],
            end_time=config["EndTime"],
        )

    elif fmt == "latlon":
        if config["StartTime"] == "NA" or config["EndTime"] == "NA":
            raise ValueError(
                "'NA' as start or end time is not supported in 'global' mode. Please "
                "specify a start and end time."
            )

        data = read_forcing_data_global(
            global_data_dir=Path(config["ForcingPath"]),
            lat=loc[0],  # type: ignore
            lon=loc[1],  # type: ignore
            start_time=np.datetime64(config["StartTime"]),
            end_time=np.datetime64(config["EndTime"]),
        )
    else:
        raise NotImplementedError

    # Write the single-column ascii '.dat' files to the input directory
    write_dat_files(data, input_path)

    # Write the two-column LAI_.dat file to the input directory.
    write_lai_file(data, input_path / "LAI_.dat")

    # Write the multi-column Mdata.txt ascii file to the input directory
    write_meteo_file(data, input_path / "Mdata.txt")

    # Write the remaining variables (without time dependency) to the matlab v7.3
    #  file 'forcing_globals.mat'
    prepare_global_variables(data, input_path)


def _slice_forcing_file(
    ds_forcing: xr.Dataset, start_time: str, end_time: str
) -> xr.Dataset:
    """Get the subset of forcing file based on time range in config.

    Also check if the desired time range is covered by forcing file.

    Args:
        ds_forcing: Dataset of forcing file.
        start_time: Start of time range in ISO format string e.g. 'YYYY-MM-DDTHH:MM:SS'.
            If "NA", start time will be the first timestamp of the forcing input data.
        end_time: End of time range in ISO format string e.g. 'YYYY-MM-DDTHH:MM:SS'.
            If "NA", end time will be the last timestamp of the forcing input data.

    Returns:
        Forcing dataset, sliced with the start and end time.
    """
    start_dtime = None if start_time == "NA" else np.datetime64(start_time)
    end_dtime = None if end_time == "NA" else np.datetime64(end_time)

    start_time_forcing = ds_forcing.coords["time"].values[0]
    end_time_forcing = ds_forcing.coords["time"].values[-1]

    start_time_valid = start_dtime >= start_time_forcing if start_dtime else True
    end_time_valid = end_dtime <= end_time_forcing if end_dtime else True
    if not (start_time_valid and end_time_valid):
        raise ValueError(
            f"Given time range (from {start_dtime} to {end_dtime}) cannot be covered by"
            f"the time range of forcing file (from {start_time_forcing} to "
            f"{end_time_forcing})."
        )

    return ds_forcing.sel(time=slice(start_dtime, end_dtime))



"""ERA5 data module, for data validation and data loading."""
from pathlib import Path
from typing import Literal
from typing import Union
import numpy as np
import PyStemmusScope.variable_conversion as vc
import xarray as xr
from PyStemmusScope.global_data import utils


ERA5_VARIABLES = ["u10", "v10", "mtpr", "sp", "ssrd", "strd"]
ERA5LAND_VARIABLES = ["t2m", "d2m"]
RESOLUTION_ERA5 = 0.25  # Resolution in degrees, to find nearest gridpoint.
RESOLUTION_ERA5LAND = 0.10


def retrieve_era5_data(
    global_data_dir: Path,
    latlon: Union[tuple[int, int], tuple[float, float]],
    time_range: tuple[np.datetime64, np.datetime64],
    timestep: str,
) -> dict:
    """Check for availability and retrieve the ERA5 and ERA5-land data.

    Args:
        global_data_dir: Path to the directory containing the global datasets.
        latlon: Latitude and longitude of the site.
        time_range: Start and end time of the model run.
        timestep: Desired timestep of the model, this is derived from the forcing data.
            In a pandas-timedelta compatible format. For example: "1800s"

    Returns:
        Dictionary containing the variables extracted from ERA5.
    """
    files_era5 = list((global_data_dir / "era5").glob("*.nc"))
    files_era5_land = list((global_data_dir / "era5-land").glob("*.nc"))

    if len(files_era5) == 0:
        raise FileNotFoundError(
            f"No netCDF files found in the folder '{global_data_dir / 'era5'}'"
        )
    if len(files_era5_land) == 0:
        raise FileNotFoundError(
            f"No netCDF files found in the folder '{global_data_dir / 'era5-land'}'"
        )

    return load_era5_data(
        files_era5,
        files_era5_land,
        latlon,
        time_range,
        timestep,
    )


def load_era5_data(
    files_era5: list[Path],
    files_era5_land: list[Path],
    latlon: Union[tuple[int, int], tuple[float, float]],
    time_range: tuple[np.datetime64, np.datetime64],
    timestep: str,
) -> dict:
    """Extract and convert the required variables from the ERA5 data.

    Args:
        files_era5: List of ERA5 files.
        files_era5_land: List of ERA5-land files.
        latlon: Latitude and longitude of the site.
        time_range: Start and end time of the model run.
        timestep: Desired timestep of the model, this is derived from the forcing data.
            In a pandas-timedelta compatible format. For example: "1800s"

    Returns:
        Dictionary containing the variables extracted from ERA5.
    """
    ds = xr.merge(
        [
            get_era5_dataset(
                files=files,
                name=name,  # type: ignore
                latlon=latlon,
                time_range=time_range,
                timestep=timestep,
            )
            for (name, files) in [("ERA5", files_era5), ("ERA5-land", files_era5_land)]
        ]
    )

    data = {}
    data["wind_speed"] = (ds["u10"] ** 2 + ds["v10"] ** 2) ** 0.5
    data["t_air_celcius"] = ds["t2m"] - 273.15  # K -> degC
    data["precip_conv"] = ds["mtpr"] / 10  # mm/s -> cm/s
    data["psurf_hpa"] = ds["sp"] / 100  # Pa -> hPa
    data["sw_down"] = ds["ssrd"] / 3600  # J * hr / m2 ->  W / m2
    data["lw_down"] = ds["strd"] / 3600  # J * hr / m2 ->  W / m2

    data["ea"] = vc.calculate_es(ds["d2m"] - 273.15) * 10  # hPa
    data["vpd"] = vc.calculate_es(data["t_air_celcius"]) * 10 - data["ea"]  # hPa
    data["rh"] = data["ea"] / vc.calculate_es(data["t_air_celcius"]) * 10
    data["Qair"] = vc.specific_humidity(data["ea"], data["psurf_hpa"])

    return data


def get_era5_dataset(
    files: list[Path],
    name: Literal["ERA5", "ERA5-land"],
    latlon: Union[tuple[int, int], tuple[float, float]],
    time_range: tuple[np.datetime64, np.datetime64],
    timestep: str,
) -> xr.Dataset:
    """Load the ERA5/ERA5-land multifile dataset, and select the location before merge.

    Args:
        files: The ERA5 or ERA5-land files.
        latlon: Latitude and longitude of the site.
        name: Either "ERA5" or "ERA5-land".
        time_range: Start and end time of the model run.
        timestep: Desired timestep of the model, this is derived from the forcing data.
            In a pandas-timedelta compatible format. For example: "1800s"

    Returns:
        The ERA5 or ERA5-land dataset.
    """
    tol = RESOLUTION_ERA5 if name == "ERA5" else RESOLUTION_ERA5LAND

    ds = xr.open_mfdataset(files, chunks="auto")

    check_era5_dataset(ds, name, latlon, time_range)

    try:
        ds = ds.sel(
            latitude=latlon[0],
            longitude=latlon[1],
            method="nearest",
            tolerance=tol,
        )
    except KeyError as err:
        if "not all values found" in str(err):
            raise utils.MissingDataError(
                f"No data point was found within {tol} degrees of the specified "
                f"location {latlon}. Please check the netCDF files or select a "
                "different location."
            ) from err
        else:
            raise err

    ds = ds.drop_vars(["latitude", "longitude"])
    ds = ds.compute()
    ds = ds.resample(time=timestep).interpolate("linear")
    return ds.sel(time=slice(time_range[0], time_range[1]))


def check_era5_dataset(
    era5data: xr.Dataset,
    name: Literal["ERA5", "ERA5-land"],
    latlon: Union[tuple[int, int], tuple[float, float]],
    time_range: tuple[np.datetime64, np.datetime64],
) -> None:
    """Validate the ERA5 or ERA5-land dataset (variables, location & time range).

    Args:
        era5data: Dataset containing the ERA5 and ERA5-land data.
        name: Either "ERA5" or "ERA5-land".
        latlon: Latitude and longitude of the site.
        time_range: Start and end time of the model run.
    """
    vars = ERA5_VARIABLES if name == "ERA5" else ERA5LAND_VARIABLES
    try:
        utils.assert_variables_present(era5data, vars)
    except utils.MissingDataError as err:
        raise utils.MissingDataError(
            f"(Some) {name} variables are missing. "
            f"Please check the {name} folder and netCDF files."
        ) from err

    try:
        utils.assert_location_within_bounds(
            era5data, x=latlon[1], y=latlon[0], xdim="longitude", ydim="latitude"
        )
    except utils.MissingDataError as err:
        raise utils.MissingDataError(
            f"The {name} data does not cover the given location. Please "
            f"check the {name} netCDF files, or select a different location."
        ) from err

    try:
        utils.assert_time_within_bounds(era5data, time_range[0], time_range[1])
    except utils.MissingDataError as err:
        raise utils.MissingDataError(
            f"The {name} data does not cover the given start and end time. "
            f"Please check the {name} netCDF files, or modify the model "
            "start and end time."
        ) from err



"""Variable conversion definitions."""
from typing import Union
import numpy as np
import xarray as xr


AVG_DENSITY_AIR = 1.292  # [kg/m3]


def calculate_ea(t_air_celcius, rh):
    """Calculate the actual vapor pressure.

    Calculate the actual vapor pressure (kPa) from the air temperature (degree Celcius)
    and relative humidity (%).

    Args:
        t_air_celcius: the air temperature in degrees C.

        rh: the relative humidity (same shape as t_air_celcius) as a percentage
            (e.g. ranging from 0 - 100).

    Returns:
        the actual vapor pressure
    """
    # Teten O. Über einige meteorologische Begriffe. Z. Geophys., 1930; 6. 297-309.
    # Murray FW. On the computation of saturation vapor pressure,
    #   J. Appl. Meteorol., 1967; 6, 203-204

    # check rh values
    if rh.min() < 0.0 or rh.max() > 100.0:
        raise ValueError(
            "relative humidity should be in percentage ranging from 0 - 100"
        )

    # check lengths
    if rh.shape != t_air_celcius.shape:
        raise ValueError("input arrays should have the same shape (size).")

    return calculate_es(t_air_celcius) * rh / 100


def calculate_es(t_celcius):
    """Calculate the saturation vapor pressure (kPa) from temperature (deg C).

    Args:
        t_celcius: the temperature in degrees C.

    Returns:
        saturation vapor pressure
    """
    # Teten O. Über einige meteorologische Begriffe. Z. Geophys., 1930; 6. 297-309.
    # Murray FW. On the computation of saturation vapor pressure,
    #   J. Appl. Meteorol., 1967; 6, 203-204

    return 0.61078 * 10 ** (t_celcius * 7.5 / (237.3 + t_celcius))


def specific_humidity(e_a, p_air):
    """Calculate the humidity [kg water / m3 air] using e_a and the air pressure.

    See: Pal Arya, S.: Introduction to Micrometeorology, Academic Press,
        San Diego, California, 1988.

    Args:
        e_a: Actual vapor pressure
        p_air: Air pressure (same units as e_a)

    Returns:
        Specific humidity [kg water / m3 air]
    """
    EPSILON = 0.622  # ratio of molecular mass of water vapour to dry air
    return EPSILON * e_a / p_air


def co2_molar_fraction_to_kg_per_m3(
    molar_fraction: Union[float, np.ndarray, xr.DataArray]
):
    """Convert CO2 molar fraction [mol cO2/mol air] to concentration in [kg CO2/m3 air].

    Note: the density of air [kg/m3] used for the calculation is assumed to be constant
    here, but will vary depending on the air pressure and air temperature.

    Args:
        molar_fraction (float, np.array): CO2 concentration as molar fraction

    Returns:
        Same as input: CO2 concentration in [kg CO2 / m3 air]
    """
    molecular_weight_co2 = 44.01  # [kg/mol]
    avg_molar_mass_air = 28.9647  # [kg/mol]
    molar_density_air = AVG_DENSITY_AIR / avg_molar_mass_air  # [mol/m3]
    #   [mol cO2/mol air] * [kg CO2/mol CO2]     * [mol air / m3]    = [kg CO2/m3 air]
    return molar_fraction * molecular_weight_co2 * molar_density_air


def co2_mass_fraction_to_kg_per_m3(
    mass_fraction: Union[float, np.ndarray, xr.DataArray]
):
    """Convert CO2 mass fraction [kg cO2/kg air] to concentration in [kg CO2/m3 air].

    Note: the density of air [kg/m3] used for the calculation is assumed to be constant
    here, but will vary depending on the air pressure and air temperature.

    Args:
        mass_fraction: CO2 concentration as mass fraction

    Returns:
        Same as input: CO2 concentration in [kg CO2 / m3 air]
    """
    return mass_fraction * AVG_DENSITY_AIR


def mask_data(data, min_value=None, max_value=None):
    """Apply a mask to data.

    Will clip the data to the min_value and max_value.

    Args:
        data (np.array): Array containing the original data.
        min_value: Minumum value of the mask.
        max_value: Maximum value of the mask

    Returns:
        np.array: Array where the values under the mask have been replaced with the
            given value.
    """
    return np.clip(data, min_value, max_value)


def field_moisture_content(theta_r, theta_s, alpha, coef_n):
    """Calculate the field moisture content at the field capacity.

    Based on the Van Genuchten equation.

    See:
        Genuchten, V. , & Th., M. . (1980). A closed-form equation for predicting the
        hydraulic conductivity of unsaturated soils. Soil Science Society of America
        Journal, 44(5), 892-898.

    Args:
        theta_r (float or np.array): Residual water content of the soil
        theta_s (float or np.array): Saturated water content of the soil
        alpha (float or np.array): Parameter related to the inverse of the air entry suction
        coef_n (float or np.array): Pore-size distribution coefficient

    Returns:
        float or np.array: Field moisture content
    """
    phi_fc = 341.9  # soil water potential at field capacity (cm)

    field_moisture_content = theta_r + (theta_s - theta_r) / (
        1 + (alpha * phi_fc) ** coef_n
    ) ** (1 - 1 / coef_n)

    return field_moisture_content


def soil_moisture(
    volumetric_water_content: np.ndarray, thickness: np.ndarray
) -> np.ndarray:
    """Calculate the soil moisture (kg/m2) from volumetric water content(m3/m3).

    based on SM = VolumetricWaterContent * Density * Thickness.

    Args:
        volumetric_water_content(np.array): volumetric water content in m3/m3
        thickness(np.array): soil layer thickness in m

    Returns:
        np.array: soil moisture in kg/m2
    """
    # check lengths
    if volumetric_water_content.shape != thickness.shape:
        raise ValueError("input arrays should have the same shape (size).")

    # Density: constant (water_density = 1000 kg per m3)
    return 1000.0 * volumetric_water_content * thickness



"""The local STEMMUS_SCOPE model process wrapper."""
import os
import platform
import subprocess
from pathlib import Path
from time import sleep
from typing import Union
from PyStemmusScope.bmi.utils import MATLAB_ERROR
from PyStemmusScope.bmi.utils import PROCESS_READY
from PyStemmusScope.bmi.utils import MatlabError
from PyStemmusScope.config_io import read_config


def alive_process(
    process: Union[subprocess.Popen, None]
) -> subprocess.Popen:  # pragma: no cover
    """Return process if the process is alive, raise an exception if it is not."""
    if process is None:
        msg = "Model process does not seem to be open."
        raise ConnectionError(msg)
    if process.poll() is not None:
        msg = f"Model terminated with return code {process.poll()}"
        raise ConnectionError(msg)
    return process


def read_stdout(process: subprocess.Popen) -> bytes:  # pragma: no cover
    """Read from stdout. If the stream ends unexpectedly, an error is raised."""
    assert process.stdout is not None  # required for type narrowing.
    read = process.stdout.read(1)

    retries = 0
    retry_time = 0.1
    while read is None:
        sleep(retry_time)
        read = process.stdout.read(1)
        retries += 1
        if retries > int(60 / retry_time):
            msg = "Connection error: could not find expected output or "
            raise ConnectionError(msg)
    return bytes(read)


def _model_is_ready(process: subprocess.Popen) -> None:  # pragma: no cover
    return _wait_for_model(PROCESS_READY, process)


def _wait_for_model(
    phrase: bytes, process: subprocess.Popen
) -> None:  # pragma: no cover
    """Wait for model to be ready for interaction."""
    output = b""

    while alive_process(process) and phrase not in output:
        output += read_stdout(process)
        if MATLAB_ERROR in output:
            try:
                process.terminate()
            finally:
                msg = (
                    "Error encountered in Matlab.\n"
                    "Please inspect logs in the output directory"
                )
                raise MatlabError(msg)


def find_exe(config: dict) -> str:
    """Find the right path to the executable file."""
    if "ExeFilePath" in config:
        exe_file = config["ExeFilePath"]
    elif os.getenv("STEMMUS_SCOPE") is not None:
        exe_file = os.getenv("STEMMUS_SCOPE")
    else:
        msg = "No STEMMUS_SCOPE executable found."
        raise ValueError(msg)
    if not Path(exe_file).exists():
        msg = f"No file found at {exe_file}"
        raise FileNotFoundError(exe_file)
    return exe_file


class LocalStemmusScope:  # pragma: no cover
    """Communicate with the local STEMMUS_SCOPE executable file."""

    def __init__(self, cfg_file: str) -> None:
        """Initialize the process."""
        self.cfg_file = cfg_file
        config = read_config(cfg_file)

        exe_file = find_exe(config)
        args = [exe_file, cfg_file, "bmi"]

        lib_path = os.getenv("LD_LIBRARY_PATH")
        if lib_path is None:
            msg = (
                "Environment variable LD_LIBRARY_PATH not found. "
                "Refer the Matlab Compiler Runtime documentation"
            )
            raise ValueError(msg)

        # Ensure output directory exists so log file can be written:
        Path(config["OutputPath"]).mkdir(parents=True, exist_ok=True)
        env = {
            "LD_LIBRARY_PATH": lib_path,
            "MATLAB_LOG_DIR": str(config["OutputPath"]),
        }

        self.process = subprocess.Popen(
            args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            bufsize=0,
            env=env,
        )

        if platform.system() == "Linux":
            assert self.process.stdout is not None  # required for type narrowing.
            # Make the connection non-blocking to allow for a timeout on read.
            os.set_blocking(self.process.stdout.fileno(), False)
        else:
            msg = "Unexpected system. The executable is only compiled for Linux."
            raise ValueError(msg)
        _model_is_ready(self.process)

    def is_alive(self) -> bool:
        """Return if the process is alive."""
        try:
            alive_process(self.process)
            return True
        except ConnectionError:
            return False

    def initialize(self) -> None:
        """Initialize the model and wait for it to be ready."""
        self.process = alive_process(self.process)

        self.process.stdin.write(  # type: ignore
            bytes(f'initialize "{self.cfg_file}"\n', encoding="utf-8")
        )
        _model_is_ready(self.process)

    def update(self) -> None:
        """Update the model and wait for it to be ready."""
        if self.process is None:
            msg = "Run initialize before trying to update the model."
            raise AttributeError(msg)

        self.process = alive_process(self.process)
        self.process.stdin.write(b"update\n")  # type: ignore
        _model_is_ready(self.process)

    def finalize(self) -> None:
        """Finalize the model."""
        self.process = alive_process(self.process)
        self.process.stdin.write(b"finalize\n")  # type: ignore
        sleep(10)
        if self.process.poll() != 0:
            try:
                self.process.terminate()
            finally:
                msg = f"Model terminated with return code {self.process.poll()}"
                raise ValueError(msg)



"""Utility funtions for the global data IO."""
from typing import Union
import numpy as np
import xarray as xr


class MissingDataError(Exception):
    """Error to be raised when requested data is missing."""


class InvalidLocationError(Exception):
    """Error to be raised when a location is given where no data will ever exist."""


def assert_variables_present(
    dataset: xr.Dataset,
    variables: list[str],
):
    """Check if the required variables are in the specified dataset."""
    dataset_variables = [str(var) for var in dataset.data_vars]
    missing_variables = set(variables) - set(dataset_variables)
    if len(missing_variables) > 0:
        raise MissingDataError(
            "Some required variables are missing from the data:"
            "".join(f"\n    {var}" for var in missing_variables)
        )


def assert_location_within_bounds(
    data: Union[xr.DataArray, xr.Dataset],
    x: float,
    y: float,
    xdim: str = "x",
    ydim: str = "y",
) -> None:
    """Compare a locations x/y (lon/lat) values to the range of the available data."""
    if (
        x > data[xdim].max()
        or x < data[xdim].min()
        or y > data[ydim].max()
        or y < data[ydim].min()
    ):
        raise MissingDataError(
            f"\nThe specified location {xdim}={x}, {ydim}={y} is not covered by the \n"
            f" range of the available data:\n"
            f"    {xdim}=[{data[xdim].min()}-{data[xdim].max()}],\n"
            f"    {ydim}=[{data[ydim].min()}-{data[ydim].max()}].\n"
        )


def assert_time_within_bounds(
    data: Union[xr.DataArray, xr.Dataset],
    start_time: np.datetime64,
    end_time: np.datetime64,
    time_dim: str = "time",
) -> None:
    """Assert that the start and end time are within the data's time bounds.

    Args:
        data: Dataset or DataArray that requires validation
        start_time: Start time of the model run.
        end_time: End time of the model run.
        time_dim: Name of the time dimension. Defaults to "time".
    """
    if (
        datetime_to_unix(start_time) < datetime_to_unix(data[time_dim].min().values)
    ) or (datetime_to_unix(end_time) > datetime_to_unix(data[time_dim].max().values)):
        raise MissingDataError(
            "\nThe available data cannot cover the specified start and end time.\n"
            f"    Specified model time range:\n"
            f"        {np.datetime_as_string(start_time, unit='m')}"  # type: ignore
            f" - {np.datetime_as_string(end_time, unit='m')}\n"  # type: ignore
            f"    Data start: {np.datetime_as_string(data[time_dim].min(), unit='m')}\n"  # type: ignore
            f"    Data end: {np.datetime_as_string(data[time_dim].max(), unit='m')}"  # type: ignore
        )


def find_nearest_non_nan(  # noqa:PLR0913 (too many arguments)
    da: xr.DataArray,
    x: float,
    y: float,
    xdim: str = "x",
    ydim: str = "y",
    max_distance: Union[float, None] = None,
) -> xr.DataArray:
    """Extract the (Cartesian) nearest non-nan value from a DataArray.

    Args:
        da: DataArray containing the data, and the xdim and ydim as dimensions
        x: x-coordinate of interest
        y: y-coordinate of interest
        xdim: optional, to be used if the x-dimension is named "lon" or "longitude".
        ydim: optional, to be used if the y-dimension is named "lat" or "latitude".
        max_distance: Maximum distance between the specified location and the nearest
            non-nan location (same units as x and y coordinates).

    Returns:
        The input dataarray reduced to only one location
    """
    distance = ((da[xdim] - x) ** 2 + (da[ydim] - y) ** 2) ** 0.5
    distance = distance.where(~np.isnan(da), np.nan)
    if max_distance is None or distance.min() < max_distance:
        return da.isel(distance.argmin(dim=[xdim, ydim]))  # type: ignore
    raise MissingDataError(
        "No non-nan data could be found within specified the maximum distance."
    )


def make_lat_lon_strings(
    lat: Union[int, float],
    lon: Union[int, float],
    step: int = 1,
) -> tuple[str, str]:
    """Turn numeric lat and lon values into strings.

    Args:
        lat: Latitude between -90 and 90.
        lon: Longitude between -180 and 180.
        step: The size of the step in values. E.g. if step is 5, valid latitude strings
            are "N00", "N05", "N10", etc.

    Returns:
        Two strings, in the form ("N50", "E005")
    """
    if lat > 90 or lat < -90:
        raise ValueError("Latitude out of bounds (-90, 90)")
    if lon > 180 or lon < -180:
        raise ValueError("Longitude out of bounds (-180, 180)")

    lat = int(lat // step * step)
    lon = int(lon // step * step)

    latstr = str(abs(lat)).rjust(2, "0")
    lonstr = str(abs(lon)).rjust(3, "0")
    latstr = f"N{latstr}" if lat >= 0 else f"S{latstr}"
    lonstr = f"E{lonstr}" if lon >= 0 else f"W{lonstr}"

    return latstr, lonstr


def datetime_to_unix(dt):
    """Convert a numpy datetime object to unix time (seconds since 1970-01-01)."""
    return dt.astype("datetime64[s]").astype(float)



"""Assorted utils to be shared across multiple other modules."""
import os
import re
from datetime import datetime
from pathlib import Path
from typing import Optional
from typing import Union
import numpy as np


def convert_to_lsm_coordinates(lat: float, lon: float) -> tuple[int, int]:
    """Convert latitude in degrees North to NCAR's LSM coordinate system.

    NCAR's LSM coordinates consist of a grid with lat values ranging from 0 -- 360,
    where 0 is the South Pole, and lon values ranging from 0 -- 720, where 0 is the
    prime meridian. Both representing a 0.5 degree resolution.

    Args:
        lat (float): Latitude in degrees North
        lon (float): longitude in degrees East

    Returns:
        (int, int): (nearest) latitude grid coordinates
    """
    lat += 90
    lat *= 2

    lon *= 2
    if lon < 0:
        lon += 720

    return np.round(lat).astype(int), np.round(lon).astype(int)


def os_name():
    """Alias for os.name."""
    return os.name


def to_absolute_path(
    input_path: Union[str, Path],
    parent: Optional[Path] = None,
    must_be_in_parent=True,
) -> Path:
    """Parse input string as :py:class:`pathlib.Path` object.

    Args:
        input_path: Input string path that can be a relative or absolute path.
        parent: Optional parent path of the input path
        must_exist: Optional argument to check if the input path exists.
        must_be_in_parent: Optional argument to check if the input path is
            subpath of parent path

    Returns:
        The input path that is an absolute path and a :py:class:`pathlib.Path` object.
    """
    must_exist = False
    pathlike = Path(input_path)
    if parent:
        if not parent.is_absolute():
            # care for windows, see issue 22
            must_exist = os_name() == "nt"
        pathlike = parent.joinpath(pathlike)
        if must_be_in_parent:
            try:
                pathlike.relative_to(parent)
            except ValueError as exc:
                raise ValueError(
                    f"Input path {input_path} is not a subpath of parent {parent}"
                ) from exc
    else:
        # care for windows, see issue 22
        must_exist = os_name() == "nt"

    return pathlike.expanduser().resolve(strict=must_exist)


def get_forcing_file(config):
    """Get forcing file from the location."""
    location, fmt = check_location_fmt(config["Location"])
    # check if the forcing file exists for the given location(s)
    if fmt == "site":
        # get forcing file list
        forcing_filenames_list = [
            file.name for file in Path(config["ForcingPath"]).iterdir()
        ]
        forcing_file = [
            filename for filename in forcing_filenames_list if location in filename
        ]
        if not forcing_file:
            raise ValueError(
                f"Forcing file does not exist for the given site {location}."
            )
        if len(forcing_file) > 1:
            raise ValueError(
                f"Multiple forcing files exist for the given site {location}."
                + "Please check your forcing files and remove the redundant files."
            )
        forcing_file = Path(config["ForcingPath"]) / forcing_file[0]

    elif fmt == "latlon":
        raise NotImplementedError
    elif fmt == "bbox":
        raise NotImplementedError
    else:
        raise ValueError("Unknown value for site format")

    return forcing_file


# Location format aliases
LatLonFmt = tuple[float, float]
BBoxFmt = tuple[LatLonFmt, LatLonFmt]


def check_location_fmt(loc: str) -> tuple[Union[str, LatLonFmt, BBoxFmt], str]:
    """Check the format of the model location.

    Three types of format are supported:
    - Site name, e.g., "DE-Kli"
    - Latitude and longitude, e.g. "(56.4, 112.0)"
    - A rectangular bounding box, described with two opposing corners:
        ((lat1, lon1), (lat2, lon2)), e.g., "((19.5, 125.5), (20.5, 130.0))".

    Args:
        loc: Location extracted from the config file.

    Returns:
        Site name (string), location (tuple), or bounding box (tuple of tuples),
        Location format (string)
    """
    # Matches a floating-point like number. I.e., 5.23 or -2.0
    flstr = r"[+-]?\d*[\.]?\d*"

    site_pattern = r"[A-Z]{2}-([A-z]|\d){3}"
    latlon_pattern = rf"\(({flstr}),\s?({flstr})\)"
    bbox_pattern = rf"\(\(({flstr}),\s?({flstr})\),\s?\(({flstr}),\s?({flstr})\)\)"

    if re.fullmatch(site_pattern, loc):
        return loc, "site"

    re_match = re.findall(latlon_pattern, loc)
    if len(re_match) == 1:
        # TODO: add check if lat/lon are valid
        return (float(re_match[0][0]), float(re_match[0][1])), "latlon"

    re_match = re.findall(bbox_pattern, loc)
    if len(re_match) == 1:
        bbox = [float(x) for x in re_match[0]]
        # TODO: add check if bbox values are valid
        return ((bbox[0], bbox[1]), (bbox[2], bbox[3])), "bbox"

    raise ValueError(
        f"Location '{loc}' in the config file does not match expected format."
    )


def _check_lat_lon(coordinates):
    """Check if the coordinates exists."""
    raise NotImplementedError


def _check_bbox(coordinates):
    """Check if the bounding box input is valid."""
    raise NotImplementedError


def check_time_fmt(start, end):
    """Check the format of time."""
    # check if start/end time can be converted to the iso format
    if start == "NA":
        start_time = None
    else:
        start_time = datetime.strptime(start, "%Y-%m-%dT%H:%M")
    if end == "NA":
        end_time = None
    else:
        end_time = datetime.strptime(end, "%Y-%m-%dT%H:%M")

    for time in [start_time, end_time]:
        if time is not None and time.minute not in [0, 30]:
            raise ValueError(
                "Invalid time values. Due to the resolution of forcing file,"
                + " the input time should be either 0 or 30 minutes."
            )

    if (start_time and end_time) and start_time > end_time:
        raise ValueError("Invalid time range. StartTime must be earlier than EndTime.")


def remove_dates_from_header(file: Path):
    """Remove the datetime string from the .mat file header.

    MATLAB raises an error when some characters are non-UTF-8 (?), e.g. Chinese month
    names. This function removes this part of the file header to avoid this problem.

    Args:
        file (Path): Valid path to the .mat file
    """
    with file.open("rb") as f:
        data = f.read()

    # Get locations of date string in header
    start_datestring = data[:128].find(b"Created on:") + 12
    end_datestring = data[:128].find(b"HDF5") - 1

    # Rebuild the data, with the dates in the header removed
    sanitized_data = (
        data[:start_datestring]
        + b" " * len(data[start_datestring:end_datestring])
        + data[end_datestring:]
    )

    # Overwrite the old file
    with file.open(mode="wb") as f:
        f.write(sanitized_data)



"""Module load and check the Prism DEM (Digital Elevation Model) dataset."""
import gzip
from pathlib import Path
from typing import Union
import xarray as xr
from PyStemmusScope.global_data import utils


MAX_DISTANCE = 0.01  #  Maximum lat/lon distance to be considered nearby. Approx 1km.


def retrieve_dem_data(
    global_data_dir: Path,
    lat: Union[int, float],
    lon: Union[int, float],
) -> float:
    """Verify if the expected canopy height file exists, and then load the data.

    Args:
        global_data_dir: Path to the directory containing the global datasets.
        lat: Latitude of the site.
        lon: Longitude of the site.

    Returns:
        Surface elevation at the location.
    """
    filename = global_data_dir / "dem" / get_filename_dem(lat, lon)

    assert_tile_existance(filename.name.replace("_DEM.tif", ".tar"))

    if not filename.exists():
        raise FileNotFoundError(
            f"\nCould not find a file with the name '{filename.name}'"
            f"\nin the directory:"
            f"\n    {filename.parent}."
            f"\nPlease download the file, or change the global data"
            f"\ndirectory to point to the right location."
        )
    return extract_prism_dem_data(filename, lat, lon)


def extract_prism_dem_data(
    file_dem: Path,
    lat: Union[int, float],
    lon: Union[int, float],
) -> float:
    """Extract the elevation from the PRISM DEM dataset.

    Args:
        file_dem: The DEM .tiff file.
        lat: Latitude of the site.
        lon: Longitude of the site.

    Returns:
        Elevation of the location.
    """
    da = xr.open_dataarray(file_dem, engine="rasterio")
    elevation = da.sel(x=lon, y=lat, method="nearest")

    try:
        elevation = utils.find_nearest_non_nan(
            da.compute(),
            x=lon,
            y=lat,
            max_distance=MAX_DISTANCE,
        )
    except utils.MissingDataError as err:
        raise utils.MissingDataError(
            f"\nNo valid DEM data found within {MAX_DISTANCE} degrees"
            f"\nof the selected location ({lat:.3f}, {lon:.3f})."
            "\nPlease select a different (nearby) location."
        ) from err

    return elevation.values[0]


def get_filename_dem(lat: Union[int, float], lon: Union[int, float]) -> str:
    """Get the right filename for the Copernicus prism DEM dataset.

    The dataset is split up in 1 degree ^2 files. This makes the output filename, for
    example with the coordinates (52N, 4E):
        Copernicus_DSM_30_N52_00_E004_00_DEM.tif

    Note: There are no files when there is only oceans/seas in the gridcell. This
        function does not take that into account.

    Args:
        lat: Latitude of the location for which the filename should be retrieved.
        lon: Longitude of the location for which the filename should be retrieved.

    Returns:
        str: Properly formatted filename for a valid DEM file.
    """
    latstr, lonstr = utils.make_lat_lon_strings(lat, lon, step=1)
    return f"Copernicus_DSM_30_{latstr}_00_{lonstr}_00_DEM.tif"


def assert_tile_existance(filename: str) -> None:
    """Assert that a DEM tile exists with the specified filename."""
    valid_name_file = (
        Path(__file__).parent / "assets" / "dem_filenames_compressed.txt.gz"
    )

    with gzip.open(valid_name_file, "rb") as f:
        valid_filenames = f.read().decode("utf-8")

    if filename not in valid_filenames:
        raise utils.InvalidLocationError(
            "\nNo DEM data tile exists for the specified location.\n"
            "Please select a different location."
        )



"""Download CAMS data using the cdsapi."""
from pathlib import Path
import cdsapi
import certifi
import urllib3


# UID+api key is read here:
with (Path.home() / ".adsloginrc").open(encoding="utf8") as f:
    uid = f.readline().strip()
    api_key = f.readline().strip()

http = urllib3.PoolManager(cert_reqs="CERT_REQUIRED", ca_certs=certifi.where())

c = cdsapi.Client(
    url="https://ads.atmosphere.copernicus.eu/api/v2",
    key=f"{uid}:{api_key}",
    verify=True,
)

c.retrieve(
    "cams-global-ghg-reanalysis-egg4",
    {
        "format": "netcdf",
        "model_level": "60",  # surface level data
        "date": "2003-01-02/2020-12-31",
        "step": ["0", "3", "6", "9", "12", "15", "18", "21"],
        "variable": "carbon_dioxide",
    },
    "CAMS_CO2_2003-2020.nc",
)



"""Module for operations related to the 'global' datasets."""
from PyStemmusScope.global_data import cams_co2
from PyStemmusScope.global_data import cci_landcover
from PyStemmusScope.global_data import copernicus_lai
from PyStemmusScope.global_data import era5
from PyStemmusScope.global_data import eth_canopy_height
from PyStemmusScope.global_data import prism_dem
from PyStemmusScope.global_data import utils
from PyStemmusScope.global_data.global_data_selection import collect_datasets


__all__ = [
    "collect_datasets",
    "utils",
    "era5",
    "eth_canopy_height",
    "prism_dem",
    "cams_co2",
    "copernicus_lai",
    "cci_landcover",
]



"""Module to load and check the ETH Canopy Height (2020) dataset."""
import gzip
from pathlib import Path
from typing import Union
import xarray as xr
from PyStemmusScope.global_data import utils


MAX_DISTANCE = 0.01  # Maximum lat/lon distance to be considered nearby.


def retrieve_canopy_height_data(
    global_data_dir: Path,
    lat: Union[int, float],
    lon: Union[int, float],
) -> float:
    """Verify if the expected canopy height file exists, and then load the data.

    Args:
        global_data_dir: Path to the directory containing the global datasets.
        lat: Latitude of the site.
        lon: Longitude of the site.

    Returns:
        Canopy height at the location.
    """
    filename = global_data_dir / "canopy_height" / get_filename_canopy_height(lat, lon)
    assert_tile_existance(filename.name)

    if not filename.exists():
        raise FileNotFoundError(
            f"\nCould not find a file with the name '{filename.name}'"
            f"\nin the directory:"
            f"\n    {filename.parent}."
            f"\nPlease download the file, or change the global data"
            f"\ndirectory to point to the right location."
        )
    return extract_canopy_height_data(filename, lat, lon)


def extract_canopy_height_data(
    file_canopy_height: Path,
    lat: Union[int, float],
    lon: Union[int, float],
) -> float:
    """Extract and validate the (ETH) canopy height data.

    Args:
        file_canopy_height: The canopy height .tiff file.
        lat: Latitude of the site.
        lon: Longitude of the site.

    Returns:
        Canopy height at the location.
    """
    da = xr.open_dataarray(file_canopy_height, engine="rasterio")
    da = da.sortby(["x", "y"])

    pad = 0.05  # Add padding around the data before trying to find nearest non-nan
    da = da.sel(y=slice(lat - pad, lat + pad), x=slice(lon - pad, lon + pad))

    try:
        canopy_height = utils.find_nearest_non_nan(
            da.compute(),
            x=lon,
            y=lat,
            max_distance=MAX_DISTANCE,
        )
    except utils.MissingDataError as err:
        raise utils.MissingDataError(
            f"\nNo valid canopy height data found within {MAX_DISTANCE} degrees"
            f"\nof the selected location ({lat:.3f}, {lon:.3f})."
            "\nPlease select a different (nearby) location."
        ) from err

    return canopy_height.values[0]


def get_filename_canopy_height(lat: Union[int, float], lon: Union[int, float]) -> str:
    """Get the right filename for the ETH canopy height dataset.

    The dataset is split up in 3 degree ^2 files. This makes the output filename, for
    example with the coordinates (52N, 4E):
        ETH_GlobalCanopyHeight_10m_2020_N51E003_Map.tif

    Note: There are no files when there is only oceans/seas in the gridcell. This
        function does not take that into account.

    Args:
        lat: Latitude of the location for which the filename should be retrieved.
        lon: Longitude of the location for which the filename should be retrieved.

    Returns:
        str: Properly formatted filename for a valid ETH GlobalCanopyHeight file.
    """
    latstr, lonstr = utils.make_lat_lon_strings(lat, lon, step=3)
    return f"ETH_GlobalCanopyHeight_10m_2020_{latstr}{lonstr}_Map.tif"


def assert_tile_existance(filename: str) -> None:
    """Assert that a canopy height tile exists with the specified filename."""
    valid_name_file = (
        Path(__file__).parent / "assets" / "h_canopy_filenames_compressed.txt.gz"
    )

    with gzip.open(valid_name_file, "rb") as f:
        valid_filenames = f.read().decode("utf-8")

    if filename not in valid_filenames:
        raise utils.InvalidLocationError(
            "\nNo canopy height data tile exists for the specified location."
            "\nPlease select a different location."
        )



"""Module for loading and validating the Copernicus LAI dataset."""
from pathlib import Path
from typing import Union
import numpy as np
import xarray as xr
from PyStemmusScope.global_data import utils


RESOLUTION_LAI = 1 / 112  # Resolution of the LAI dataset in degrees


def retrieve_lai_data(
    global_data_dir: Path,
    latlon: Union[tuple[int, int], tuple[float, float]],
    time_range: tuple[np.datetime64, np.datetime64],
    timestep: str,
) -> np.ndarray:
    """Check for availability and retrieve the Copernicus LAI data.

    Args:
        global_data_dir: Path to the directory containing the global datasets.
        latlon: Latitude and longitude of the site.
        time_range: Start and end time of the model run.
        timestep: Desired timestep of the model, this is derived from the forcing data.
            In a pandas-timedelta compatible format. For example: "1800s"

    Returns:
        DataArray containing the LAI of the specified site for the given time range.
    """
    files = list((global_data_dir / "lai").glob("*.nc"))

    if len(files) == 0:
        raise FileNotFoundError(
            f"No netCDF files found in the folder '{global_data_dir / 'lai'}'"
        )

    return extract_lai_data(
        files_lai=files,
        latlon=latlon,
        time_range=time_range,
        timestep=timestep,
    )


def extract_lai_data(
    files_lai: list[Path],
    latlon: Union[tuple[int, int], tuple[float, float]],
    time_range: tuple[np.datetime64, np.datetime64],
    timestep: str,
) -> np.ndarray:
    """Generate LAI values, until a dataset is chosen.

    Args:
        files_lai: List of paths to the *.nc files.
        latlon: Latitude and longitude of the site.
        time_range: Start and end time of the model run.
        timestep: Desired timestep of the model, this is derived from the forcing data.
            In a pandas-timedelta compatible format. For example: "1800s"

    Returns:
        DataArray containing the LAI of the specified site for the given time range.
    """
    ds = xr.open_mfdataset(files_lai, chunks="auto")

    check_lai_dataset(ds, latlon, time_range)

    ds = ds.drop_vars(["crs", "LAI_ERR", "retrieval_flag"])

    try:
        ds = ds.sel(
            lat=latlon[0],
            lon=latlon[1],
            method="nearest",
            tolerance=RESOLUTION_LAI,
        )
    except KeyError as err:
        if "not all values found" in str(err):
            raise utils.MissingDataError(
                f"\nNo data point was found within {RESOLUTION_LAI} degrees"
                f"\nof the specified location {latlon}."
                f"\nPlease check the netCDF files or select a different location"
            ) from err
        else:
            raise err

    ds = ds.drop_vars(["lat", "lon"])
    ds = ds.compute()  # Load into memory before resampling
    ds = ds.resample(time=timestep).interpolate("linear")
    ds = ds.sel(time=slice(time_range[0], time_range[1]))

    return ds["LAI"].values


def check_lai_dataset(
    lai_data: xr.Dataset,
    latlon: Union[tuple[int, int], tuple[float, float]],
    time_range: tuple[np.datetime64, np.datetime64],
) -> None:
    """Validate the LAI dataset (variables name, location & time range).

    Args:
        lai_data: Dataset containing the LAI data.
        latlon: Latitude and longitude of the site.
        time_range: Start and end time of the model run.
    """
    try:
        utils.assert_variables_present(lai_data, ["LAI"])
    except utils.MissingDataError as err:
        raise utils.MissingDataError(
            "\nCould not find the variable 'LAI' in the LAI dataset. "
            "\nPlease check the netCDF files or the path."
        ) from err

    try:
        utils.assert_location_within_bounds(
            lai_data, x=latlon[1], y=latlon[0], xdim="lon", ydim="lat"
        )
    except utils.MissingDataError as err:
        raise utils.MissingDataError(
            "\nThe LAI data does not cover the given location."
            "\nPlease check the LAI netCDF files, or select a different location."
        ) from err

    try:
        utils.assert_time_within_bounds(lai_data, time_range[0], time_range[1])
    except utils.MissingDataError as err:
        raise utils.MissingDataError(
            "\nThe LAI data does not cover the given start and end time. "
            "\nPlease check the LAI netCDF files, or modify the model"
            "\nstart and end time."
        ) from err



"""PyStemmusScope directories utilities.

Module designed to manage input directories and data for running the model
and storing outputs.
"""
import logging
import shutil
import time
from pathlib import Path
from typing import Union
from . import utils


logger = logging.getLogger(__name__)


def read_config(config_file: Union[str, Path]) -> dict[str, str]:
    """Read config from given config file.

    Load paths from config file and save them into dict.

    Args:
        config_file: Path to the config file.

    Returns:
        Dictionary containing paths to work directory and all sub-directories.
    """
    config = {}
    with Path(config_file).open(encoding="utf8") as f:
        for line in f:
            (key, val) = line.split("=")
            config[key] = val.rstrip("\n")

    validate_config(config)

    return config


def validate_config(config: Union[Path, dict]):
    """Validate the config file."""
    if isinstance(config, dict):
        cfg = config  # For proper type narrowing understood by Mypy.
    elif isinstance(config, Path):
        cfg = read_config(config)
    else:
        raise ValueError(
            "The input to validate_config should be either a Path or dict"
            f" object, but a {type(config)} object was passed."
        )

    # TODO: add check if the input data directories/file exist, and return clear error to user.
    _ = utils.check_location_fmt(cfg["Location"])
    utils.check_time_fmt(cfg["StartTime"], cfg["EndTime"])


def create_io_dir(config: dict) -> tuple[Path, Path, Path]:
    """Create input directory and copy required files.

    Work flow executor to create work directory and all sub-directories.

    Returns:
        Path to input, output directory and config file for every station/forcing.
    """
    # get start time with the format Y-M-D-HM
    timestamp = time.strftime("%Y-%m-%d-%H%M")

    loc, fmt = utils.check_location_fmt(config["Location"])
    if fmt == "site":
        site_name = loc
        input_dir_name = f"{loc}_{timestamp}"
    elif fmt == "latlon":
        site_name = "global"
        latstr = f"{loc[0]:.3f}".replace(".", "-")
        lonstr = f"{loc[1]:.3f}".replace(".", "-")
        latstr = f"N{latstr}" if loc[0] >= 0 else f"S{latstr[1:]}"  # type: ignore
        lonstr = f"E{lonstr}" if loc[1] >= 0 else f"W{lonstr[1:]}"  # type: ignore
        input_dir_name = f"global_{latstr}_{lonstr}_{timestamp}"
    else:
        raise NotImplementedError()

    # create input directory
    work_dir = utils.to_absolute_path(config["WorkDir"])
    input_dir = work_dir / "input" / input_dir_name
    input_dir.mkdir(parents=True, exist_ok=True)
    message = f"Prepare work directory {input_dir} for the location: {loc}"
    logger.info("%s", message)

    # copy model parameters to work directory
    _copy_data(input_dir, config)

    # create output directory
    output_dir = work_dir / "output" / input_dir_name
    output_dir.mkdir(parents=True, exist_ok=True)
    message = f"Prepare work directory {output_dir} for the location: {loc}"
    logger.info("%s", message)

    # update config file for ForcingFileName and InputPath
    config_file_path = _update_config_file(
        input_dir, output_dir, config, site_name, timestamp  # type: ignore
    )

    return input_dir, output_dir, config_file_path


def _copy_data(input_dir: Path, config: dict) -> None:
    """Copy required data to the work directory.

    Create sub-directories inside the work directory and copy data.

    Args:
        input_dir: Path to the input directory.
        config: Dictionary containing all the paths.
    """
    folder_list_vegetation = [
        "directional",
        "fluspect_parameters",
        "leafangles",
        "radiationdata",
        "soil_spectrum",
    ]
    for folder in folder_list_vegetation:
        (input_dir / folder).mkdir(parents=True, exist_ok=True)
        shutil.copytree(
            str(config[folder]), str(input_dir / folder), dirs_exist_ok=True
        )

    # copy input_data.xlsx
    shutil.copy(str(config["input_data"]), str(input_dir))


def _update_config_file(
    input_dir: Path,
    output_dir: Path,
    config: dict,
    site_name: str,
    timestamp: str,
) -> Path:
    """Update config file for each station.

    Create config file for each forcing/station under the work directory.

    Args:
        input_dir: Path to the input directory.
        output_dir: Path to the output directory.
        config: Dictionary containing all the paths.
        site_name: The site name (eg. "FI-Hyy"), or "latlon" for global data.
        timestamp: Timestamp when creating the config file.

    Returns:
        Path to updated config file.
    """
    config_file_path = input_dir / f"{site_name}_{timestamp}_config.txt"
    with config_file_path.open(mode="w", encoding="utf8") as f:
        for key, value in config.items():
            if key == "InputPath":
                update_entry = f"{key}={str(input_dir)}/\n"
            elif key == "OutputPath":
                update_entry = f"{key}={str(output_dir)}/\n"
            else:
                update_entry = f"{key}={value}\n"

            f.write(update_entry)

    return config_file_path



"""Module for loading and validating the ESA CCI land cover dataset."""
from pathlib import Path
from typing import Union
import numpy as np
import pandas as pd
import xarray as xr
from PyStemmusScope.global_data import utils


RESOLUTION_CCI = 1 / 360  # Resolution of the dataset in degrees
FILEPATH_LANDCOVER_TABLE = Path(__file__).parent / "assets" / "lccs_to_igbp_table.csv"


def retrieve_landcover_data(
    global_data_dir: Path,
    latlon: Union[tuple[int, int], tuple[float, float]],
    time_range: tuple[np.datetime64, np.datetime64],
    timestep: str,
) -> dict[str, np.ndarray]:
    """Get the land cover data from the CCI netCDF files.

    Args:
        global_data_dir: Path to the directory containing the global datasets.
        latlon: Latitude and longitude of the site.
        time_range: Start and end time of the model run.
        timestep: Desired timestep of the model, this is derived from the forcing data.
            In a pandas-timedelta compatible format. For example: "1800s"

    Returns:
        Dictionary containing IGBP and LCCS land cover classes.
    """
    files_cci = list((global_data_dir / "landcover").glob("*.nc"))

    if len(files_cci) == 0:
        raise FileNotFoundError(
            f"No netCDF files found in the folder '{global_data_dir / 'landcover'}'"
        )

    return extract_landcover_data(
        files_cci=files_cci,
        latlon=latlon,
        time_range=time_range,
        timestep=timestep,
    )


def extract_landcover_data(
    files_cci: list[Path],
    latlon: Union[tuple[int, int], tuple[float, float]],
    time_range: tuple[np.datetime64, np.datetime64],
    timestep: str,
) -> dict[str, np.ndarray]:
    """Extract the land cover data from the CCI netCDF files.

    Args:
        files_cci: List of CCI land cover files.
        latlon: Latitude and longitude of the site.
        time_range: Start and end time of the model run.
        timestep: Desired timestep of the model, this is derived from the forcing data.
            In a pandas-timedelta compatible format. For example: "1800s"

    Returns:
        Dictionary containing IGBP and LCCS land cover classes.
    """
    cci_dataset = xr.open_mfdataset(files_cci, chunks="auto")

    check_cci_dataset(cci_dataset, latlon, time_range)  # Assert spatial/temporal bounds

    lat_bounds = cci_dataset["lat_bounds"].load()  # Load so that they are not
    lon_bounds = cci_dataset["lon_bounds"].load()  #  dask arrays
    lat_idx = np.logical_and(  # type: ignore
        lat_bounds.isel(bounds=0) >= latlon[0], lat_bounds.isel(bounds=1) < latlon[0]
    ).argmax(dim="lat")
    lon_idx = np.logical_and(  # type: ignore
        lon_bounds.isel(bounds=0) <= latlon[1], lon_bounds.isel(bounds=1) > latlon[1]
    ).argmax(dim="lon")

    lccs_id = cci_dataset.isel(lat=lat_idx, lon=lon_idx)["lccs_class"]

    # If time is size 1, interp fails. Adding an extra datapoint prevents this.
    if lccs_id["time"].size == 1:
        data_copy = lccs_id.copy()
        data_copy["time"] = lccs_id["time"] + np.timedelta64(1, "D")
        lccs_id = xr.concat((lccs_id, data_copy), dim="time")

    lccs_id = lccs_id.interp(
        time=pd.date_range(time_range[0], time_range[1], freq=timestep),
        method="nearest",
        kwargs={"fill_value": "extrapolate", "bounds_error": False},
    )

    landcover_lookup_table = get_landcover_table(cci_dataset)
    igbp_lookup_table = get_lccs_to_igbp_table()

    return {
        "LCCS_landcover": np.array(
            [landcover_lookup_table[_id] for _id in lccs_id.to_numpy()]
        ),
        "IGBP_veg_long": np.array(
            [igbp_lookup_table[_id] for _id in lccs_id.to_numpy()]
        ),
    }


def get_lccs_to_igbp_table() -> dict[int, str]:
    """Read the land cover translation table, and turn it into a lookup dictionary."""
    df = pd.read_csv(FILEPATH_LANDCOVER_TABLE, index_col="lccs_class")
    return df.to_dict()["IGBP_STEMMUS_SCOPE"]


def get_landcover_table(cci_dataset: xr.Dataset) -> dict[int, str]:
    """Get the lookup table to convert the flag values to a land cover name.

    The lookup table for the land cover classes is contained in the netCDF file, under
    the lcc_class attributes. This function extracts it and turns it into a (dict)
    lookup table.

    Args:
        cci_dataset: The CCI dataset netCDF file.

    Returns:
        The landcover class lookup table
    """
    flag_meanings = cci_dataset["lccs_class"].attrs["flag_meanings"].split(" ")
    flag_values = cci_dataset["lccs_class"].attrs["flag_values"]
    return dict(zip(flag_values, flag_meanings))


def check_cci_dataset(
    cci_dataset: xr.Dataset,
    latlon: Union[tuple[int, int], tuple[float, float]],
    time_range: tuple[np.datetime64, np.datetime64],
) -> None:
    """Validate the cci dataset for spatial and temporal bounds."""
    # Assert spatial bounds
    if (
        latlon[0] > cci_dataset["lat"].max() or latlon[0] < cci_dataset["lat"].min()
    ) or (latlon[1] > cci_dataset["lon"].max() or latlon[1] < cci_dataset["lon"].min()):
        raise utils.MissingDataError(
            f"\nThe specified location {latlon} was not within bounds of the CCI land"
            f"\ncover dataset."
            f"\nPlease check the netCDF files or select a different location"
        )

    # Assert temporal bounds
    first_year_cci = pd.to_datetime(cci_dataset["time"].min().to_numpy()).year
    last_year_cci = pd.to_datetime(cci_dataset["time"].max().to_numpy()).year
    first_year_range = pd.to_datetime(time_range[0]).year
    last_year_range = pd.to_datetime(time_range[-1]).year
    # As the data is yearly, allow some leeway with the time bounds
    if (first_year_range + 1 < first_year_cci) or (last_year_range - 1 > last_year_cci):
        raise utils.MissingDataError(
            f"\nThe specified time range {time_range} was not within the range of the"
            f"\nCCI land cover dataset:"
            f"\n({cci_dataset['time'].min(), cci_dataset['time'].max()})"
            f"\nPlease check the netCDF files or select a different location"
        )



"""BMI wrapper for the STEMMUS_SCOPE model."""
import os
from pathlib import Path
from typing import Literal
from typing import Protocol
from typing import Union
import h5py
import numpy as np
from bmipy.bmi import Bmi
from PyStemmusScope.bmi.utils import InapplicableBmiMethods
from PyStemmusScope.config_io import read_config


MODEL_INPUT_VARNAMES: tuple[str, ...] = ("soil_temperature",)

MODEL_OUTPUT_VARNAMES: tuple[str, ...] = (
    "soil_temperature",
    "respiration",
)

MODEL_VARNAMES: tuple[str, ...] = tuple(
    set(MODEL_INPUT_VARNAMES + MODEL_OUTPUT_VARNAMES)
)

VARNAME_UNITS: dict[str, str] = {"respiration": "unknown", "soil_temperature": "degC"}

VARNAME_DTYPE: dict[str, str] = {
    "respiration": "float64",
    "soil_temperature": "float64",
}

VARNAME_GRID: dict[str, int] = {
    "respiration": 0,
    "soil_temperature": 1,
}

NO_STATE_MSG = (
    "The model state is not available. Please run `.update()` before requesting "
    "\nthis model info. If you did run .update() before, something seems to have "
    "\ngone wrong and you have to restart the model."
)

NO_CONFIG_MSG = (
    "The model has not been initialized with a configuration file yet. Please first run"
    "\n.initialize() before requesting this model info."
)


def load_state(config: dict) -> h5py.File:
    """Load the STEMMUS_SCOPE model state.

    Args:
        config: BMI configuration, containing the path to the output directory.

    Returns:
        Model state, as a dict.
    """
    matfile = Path(config["OutputPath"]) / "STEMMUS_SCOPE_state.mat"
    return h5py.File(matfile, mode="a")


def get_variable(state: h5py.File, varname: str) -> np.ndarray:
    """Get a variable from the model state.

    Args:
        state: STEMMUS_SCOPE model state
        varname: Variable name
    """
    if varname == "respiration":
        return state["fluxes"]["Resp"][0]
    elif varname == "soil_temperature":
        return state["TT"][0, :-1]
    else:
        if varname in MODEL_VARNAMES:
            msg = "Varname is missing in get_variable! Contact devs."
        else:
            msg = "Unknown variable name"
        raise ValueError(msg)


def set_variable(
    state: h5py.File,
    varname: str,
    value: np.ndarray,
    inds: Union[np.ndarray, None] = None,
) -> dict:
    """Set a variable in the model state.

    Args:
        state: Model state.
        varname: Variable name.
        value: New value for the variable.
        inds: (Optional) at which indices you want to set the variable values.

    Returns:
        Updated model state.
    """
    if inds is not None:
        vals = get_variable(state, varname)
        vals[inds] = value
    else:
        vals = value

    if varname == "soil_temperature":
        state["TT"][0, :-1] = vals
    else:
        if varname in MODEL_OUTPUT_VARNAMES and varname not in MODEL_INPUT_VARNAMES:
            msg = "This variable is a model output variable only. You cannot set it."
        elif varname in MODEL_VARNAMES:
            msg = "Varname is missing in set_variable! Contact devs."
        else:
            msg = "Uknown variable name"
        raise ValueError(msg)
    return state


def get_run_mode(config: dict) -> Literal["exe", "docker"]:
    """Get the run mode (docker or EXE) from the config file.

    Args:
        config: Config dictionary

    Returns:
        Run mode (either "exe" or "docker").
    """
    if "ExeFilePath" in config:
        return "exe"
    elif "DockerImage" in config:
        return "docker"
    elif os.getenv("STEMMUS_SCOPE") is not None:
        return "exe"
    else:
        msg = (
            "No valid config found, or the STEMMUS_SCOPE environment variable is "
            "not set.\nPlease use the ExeFilePath or DockerImage configuration entry, "
            "or set the STEMMUS_SCOPE environment variable."
        )
        raise ValueError(msg)


def check_writable(file: Path) -> None:
    """Check if this process has write access to a file."""
    if not os.access(file, os.W_OK):
        msg = (
            f"The file '{file}' already exists, and this process has no"
            " write access to it."
        )
        raise PermissionError(msg)


class StemmusScopeProcess(Protocol):
    """Protocol for communicating with the model process."""

    def __init__(self, cfg_file: str) -> None:
        """Initialize the process class (e.g. create the container)."""
        ...

    def is_alive(self) -> bool:
        """Return if the process is alive."""
        ...

    def initialize(self) -> None:
        """Initialize the model and wait for it to be ready."""
        ...

    def update(self) -> None:
        """Update the model and wait for it to be ready."""
        ...

    def finalize(self) -> None:
        """Finalize the model."""
        ...


def start_process(mode: Literal["exe", "docker"], cfg_file: str) -> StemmusScopeProcess:
    """Start the right STEMMUS_SCOPE process."""
    if mode == "docker":
        try:
            from PyStemmusScope.bmi.docker_process import StemmusScopeDocker

            return StemmusScopeDocker(cfg_file=cfg_file)
        except ImportError as err:
            msg = (
                "The docker python package is not available."
                " Please install before continuing."
            )
            raise ImportError(msg) from err
    elif mode == "exe":
        from PyStemmusScope.bmi.local_process import LocalStemmusScope

        return LocalStemmusScope(cfg_file=cfg_file)
    else:
        msg = "Unknown mode."
        raise ValueError(msg)


class StemmusScopeBmi(InapplicableBmiMethods, Bmi):
    """STEMMUS_SCOPE Basic Model Interface."""

    config_file: str = ""
    config: dict = {}
    state: Union[h5py.File, None] = None
    state_file: Union[Path, None] = None

    _run_mode: Union[str, None] = None
    _process: Union[StemmusScopeProcess, None] = None

    def initialize(self, config_file: str) -> None:
        """Perform startup tasks for the model.

        Args:
            config_file: Path to the configuration file.
        """
        self.config_file = config_file
        self.config = read_config(config_file)

        Path(self.config["OutputPath"]).mkdir(parents=True, exist_ok=True)
        self.state_file = Path(self.config["OutputPath"]) / "STEMMUS_SCOPE_state.mat"
        if self.state_file.exists():
            check_writable(self.state_file)
        else:
            self.state_file.touch()  # Prevent docker messing up file permission.

        self._run_mode = get_run_mode(self.config)

        self._process = start_process(self._run_mode, config_file)
        self._process.initialize()

    def update(self) -> None:
        """Advance the model state by one time step."""
        if self.state is not None:
            self.state = self.state.close()  # Close file to allow matlab to write

        if self._process is not None:
            self._process.update()
        else:
            msg = "The STEMMUS_SCOPE process is not running/connected. Can't update!"
            raise ValueError(msg)

        self.state = load_state(self.config)

    def update_until(self, time: float) -> None:
        """Advance model state until the given time.

        Args:
            time: A model time later than the current model time.
        """
        while time > self.get_current_time():
            self.update()

    def finalize(self) -> None:
        """Finalize the STEMMUS_SCOPE model."""
        if self._process is not None:
            self._process.finalize()
        else:
            msg = "The STEMMUS_SCOPE process is not running/connected. Can't finalize!"
            raise ValueError(msg)

    def get_component_name(self) -> str:
        """Name of the component.

        Returns:
            Name of the component (STEMMUS_SCOPE).
        """
        return "STEMMUS_SCOPE"

    ### VARIABLE INFO METHODS ###
    def get_input_item_count(self) -> int:
        """Get the number of model input variables.

        Returns:
            The number of input variables.
        """
        return len(MODEL_INPUT_VARNAMES)

    def get_output_item_count(self) -> int:
        """Get the number of model output variables.

        Returns:
            The number of output variables.
        """
        return len(MODEL_OUTPUT_VARNAMES)

    # The types of the following two methods are wrong in python-bmi
    # see: https://github.com/csdms/bmi-python/issues/38
    def get_input_var_names(self) -> tuple[str, ...]:  # type: ignore
        """List of the model's input variables (as CSDMS Standard Names)."""
        return MODEL_INPUT_VARNAMES

    def get_output_var_names(self) -> tuple[str, ...]:  # type: ignore
        """List of the model's output variables (as CSDMS Standard Names)."""
        return MODEL_OUTPUT_VARNAMES

    def get_var_grid(self, name: str) -> int:
        """Get grid identifier for the given variable."""
        return VARNAME_GRID[name]

    def get_var_type(self, name: str) -> str:
        """Get data type of the given variable."""
        return VARNAME_DTYPE[name]

    def get_var_units(self, name: str) -> str:
        """Get units of the given variable."""
        return VARNAME_UNITS[name]

    def get_var_itemsize(self, name: str) -> int:
        """Get memory use for each array element in bytes."""
        return np.array([], dtype=VARNAME_DTYPE[name]).itemsize

    def get_var_nbytes(self, name: str) -> int:
        """Get size, in bytes, of the given variable."""
        return self.get_grid_size(self.get_var_grid(name)) * self.get_var_itemsize(name)

    ### TIME METHODS ###
    def get_current_time(self) -> float:
        """Get the current time of the model."""
        if self.state is None:
            raise ValueError(NO_STATE_MSG)

        return self.get_start_time() + np.sum(self.state["TimeStep"][0])

    def get_start_time(self) -> float:
        """Start time of the model."""
        if len(self.config) == 0:
            raise ValueError(NO_CONFIG_MSG)

        return (
            np.datetime64(self.config["StartTime"])
            .astype("datetime64[s]")
            .astype("float")
        )

    def get_end_time(self) -> float:
        """End time of the model."""
        if len(self.config) == 0:
            raise ValueError(NO_CONFIG_MSG)

        return (
            np.datetime64(self.config["EndTime"])
            .astype("datetime64[s]")
            .astype("float")
        )

    def get_time_units(self) -> str:
        """Time units of the model."""
        return "seconds since 1970-01-01 00:00:00.0 +0000"

    def get_time_step(self) -> float:
        """Return the current time step of the model."""
        if self.state is None:
            raise ValueError(NO_STATE_MSG)
        return float(self.state["TimeStep"][0][0])

    ### GETTERS AND SETTERS ###
    def get_value(self, name: str, dest: np.ndarray) -> np.ndarray:
        """Get a copy of values of the given variable.

        Args:
            name: input or output variable name, a CSDMS Standard Name.
            dest: numpy array into which to place the values.

        Returns:
            The same numpy array that was passed as an input buffer.
        """
        if self.state is None:
            raise ValueError(NO_STATE_MSG)
        dest[:] = get_variable(self.state, name)
        return dest

    def get_value_ptr(self, name: str) -> np.ndarray:
        """Get a reference to values of the given variable.

        Note: not possible due to the Matlab<>Python coupling.
        """
        raise NotImplementedError()

    def get_value_at_indices(
        self, name: str, dest: np.ndarray, inds: np.ndarray
    ) -> np.ndarray:
        """Get values at particular indices.

        Args:
            name: Input or output variable name, a CSDMS Standard Name.
            dest: numpy array into which to place the values.
            inds: The indices into the variable array.

        Returns:
            Value of the model variable at the given location.
        """
        if self.state is None:
            raise ValueError(NO_STATE_MSG)
        dest[:] = get_variable(self.state, name)[inds]
        return dest

    def set_value(self, name: str, src: np.ndarray) -> None:
        """Specify a new value for a model variable.

        Args:
            name: Input or output variable name, a CSDMS Standard Name.
            src: The new value for the specified variable.
        """
        if self.state is None:
            raise ValueError(NO_STATE_MSG)
        self.state = set_variable(self.state, name, src)

    def set_value_at_indices(
        self, name: str, inds: np.ndarray, src: np.ndarray
    ) -> None:
        """Specify a new value for a model variable at particular indices.

        Parameters
        ----------
        name : str
            An input or output variable name, a CSDMS Standard Name.
        inds : array_like
            The indices into the variable array.
        src : array_like
            The new value for the specified variable.
        """
        if self.state is None:
            raise ValueError(NO_STATE_MSG)
        self.state = set_variable(self.state, name, src, inds)

    ### GRID INFO ###
    def get_grid_rank(self, grid: int) -> int:
        """Get number of dimensions of the computational grid.

        Args:
            grid: A grid identifier.

        Returns:
            Rank of the grid.
        """
        if grid == 0:
            return 2
        if grid == 1:
            return 3
        msg = f"Invalid grid identifier '{grid}'"
        raise ValueError(msg)

    def get_grid_size(self, grid: int) -> int:
        """Get the total number of elements in the computational grid.

        Args:
            grid: A grid identifier.

        Returns:
            Size of the grid.
        """
        if self.state is None:
            raise ValueError(NO_STATE_MSG)

        if grid == 0:
            return 1
        if grid == 1:
            return int(self.state["ModelSettings"]["mN"][0]) - 1

        msg = f"Invalid grid identifier '{grid}'"
        raise ValueError(msg)

    def get_grid_type(self, grid: int) -> str:
        """Get the grid type as a string.

        Args:
            grid: A grid identifier.

        Returns:
            Type of grid as a string.
        """
        return "rectilinear"

    def get_grid_x(self, grid: int, x: np.ndarray) -> np.ndarray:
        """Get coordinates of grid nodes in the x direction.

        Args:
            grid: grid identifier.
            x: numpy array to hold the x-coordinates of the grid node columns.

        Returns:
            The input numpy array that holds the grid's column x-coordinates.
        """
        if self.state is None:
            raise ValueError(NO_STATE_MSG)
        x[:] = self.state["SiteProperties"]["longitude"][0]
        return x

    def get_grid_y(self, grid: int, y: np.ndarray) -> np.ndarray:
        """Get coordinates of grid nodes in the y direction.

        Args:
            grid: grid identifier.
            y: numpy array to hold the y-coordinates of the grid node columns.

        Returns:
            The input numpy array that holds the grid's column y-coordinates.
        """
        if self.state is None:
            raise ValueError(NO_STATE_MSG)
        y[:] = self.state["SiteProperties"]["latitude"][0]
        return y

    def get_grid_z(self, grid: int, z: np.ndarray) -> np.ndarray:
        """Get coordinates of grid nodes in the z direction.

        Args:
            grid: grid identifier.
            z: numpy array to hold the z-coordinates of the grid node columns.

        Returns:
            The input numpy array that holds the grid's column z-coordinates.
        """
        if self.state is None:
            raise ValueError(NO_STATE_MSG)
        if grid == 1:
            z[:] = (
                -np.hstack(
                    (
                        self.state["ModelSettings"]["DeltZ_R"][:, 0].cumsum()[::-1],
                        np.array([0.0]),
                    )
                )
                / 100
            )
            return z
        else:
            raise ValueError(f"Grid {grid} has no dimension `z`.")

    def get_grid_shape(self, grid: int, shape: np.ndarray) -> np.ndarray:
        """Get dimensions of the computational grid."""
        if grid not in [0, 1]:
            msg = f"Unknown grid identifier '{grid}'"
            raise ValueError(msg)

        shape[-1] = 1  # Last element is x
        shape[-2] = 1  # Semi-last element is y
        if grid == 1:
            shape[-3] = self.get_grid_size(grid)  # First element is z
        return shape



"""Module containing the full global dataset collection."""
from pathlib import Path
from typing import Any
from typing import Union
import numpy as np
import pandas as pd
import xarray as xr
from PyStemmusScope import global_data as gd
from PyStemmusScope import variable_conversion as vc


def collect_datasets(
    global_data_dir: Path,
    latlon: Union[tuple[int, int], tuple[float, float]],
    time_range: tuple[np.datetime64, np.datetime64],
    timestep: str,
) -> dict:
    """Collect and merge all the global datasets into one.

    Args:
        global_data_dir: Path to the directory containing the global datasets.
        latlon: Latitude and longitude of the site.
        time_range: Start and end time of the model run.
        timestep: Desired timestep of the model, this is derived from the forcing data.
            In a pandas-timedelta compatible format. For example: "1800s"

    Returns:
        Dictionary containing the variables extracted from the global datasets.
    """
    data: dict[str, Any] = {
        "time": xr.DataArray(
            pd.date_range(str(time_range[0]), str(time_range[1]), freq=timestep).rename(
                "time"
            )
        )
    }
    era5_data = gd.era5.retrieve_era5_data(
        global_data_dir,
        latlon,
        time_range,
        timestep,
    )

    data = {**data, **era5_data}

    data["wind_speed"] = vc.mask_data(data["wind_speed"], min_value=0.05)

    data["co2_conv"] = (
        vc.co2_mass_fraction_to_kg_per_m3(
            gd.cams_co2.retrieve_co2_data(
                global_data_dir,
                latlon,
                time_range,
                timestep,
            )
        )
        * 1e6
    )  # kg/m3 -> mg/m3

    data["lai"] = vc.mask_data(
        gd.copernicus_lai.retrieve_lai_data(
            global_data_dir,
            latlon,
            time_range,
            timestep,
        ),
        min_value=0.01,
    )

    data["elevation"] = gd.prism_dem.retrieve_dem_data(
        global_data_dir, latlon[0], latlon[1]
    )

    data["canopy_height"] = gd.eth_canopy_height.retrieve_canopy_height_data(
        global_data_dir,
        latlon[0],
        latlon[1],
    )

    # Height of measurement. Data has no actual equivalent. Set to a temporary value.
    # See issue #145.
    data["reference_height"] = 10.0

    data["sitename"] = "global"

    # Expected time format is days (as floating point) since Jan 1st 00:00.
    data["doy_float"] = (
        data["time"].dt.dayofyear
        - 1
        + data["time"].dt.hour / 24
        + data["time"].dt.minute / 60 / 24
    )
    data["year"] = data["time"].dt.year.astype(float)

    data["DELT"] = (
        (data["time"].values[1] - data["time"].values[0]) / np.timedelta64(1, "s")
    ).astype(float)
    data["total_timesteps"] = data["time"].size

    data["latitude"] = latlon[0]
    data["longitude"] = latlon[1]

    landcover_data = gd.cci_landcover.retrieve_landcover_data(
        global_data_dir,
        latlon,
        time_range,
        timestep,
    )
    data["IGBP_veg_long"] = landcover_data["IGBP_veg_long"]
    data["LCCS_landcover"] = landcover_data["LCCS_landcover"]

    return data



"""Utilities for the STEMMUS_SCOPE Basic Model Interface."""
import numpy as np


# Phrases defined in the Matlab code to check for:
PROCESS_READY = b"Select BMI mode:"
PROCESS_FINALIZED = b"Finished clean up."
MATLAB_ERROR = b"Error in "


INAPPLICABLE_GRID_METHOD_MSG = (
    "This grid method is not implmented for the STEMMUS_SCOPE BMI because the model is"
    "\non a rectilinear grid."
)


class MatlabError(Exception):
    """Matlab code encountered an error."""

    pass


class InapplicableBmiMethods:
    """Holds methods that are not applicable for STEMMUS_SCOPE's rectilinear grid."""

    def get_grid_spacing(self, grid: int, spacing: np.ndarray) -> np.ndarray:
        """Get distance between nodes of the computational grid."""
        raise NotImplementedError(INAPPLICABLE_GRID_METHOD_MSG)

    def get_grid_origin(self, grid: int, origin: np.ndarray) -> np.ndarray:
        """Get coordinates for the lower-left corner of the computational grid."""
        raise NotImplementedError(INAPPLICABLE_GRID_METHOD_MSG)

    def get_var_location(self, name: str) -> str:
        """Get the grid element type that the a given variable is defined on."""
        raise NotImplementedError(INAPPLICABLE_GRID_METHOD_MSG)

    def get_grid_node_count(self, grid: int) -> int:
        """Get the number of nodes in the grid."""
        raise NotImplementedError(INAPPLICABLE_GRID_METHOD_MSG)

    def get_grid_edge_count(self, grid: int) -> int:
        """Get the number of edges in the grid."""
        raise NotImplementedError(INAPPLICABLE_GRID_METHOD_MSG)

    def get_grid_face_count(self, grid: int) -> int:
        """Get the number of faces in the grid."""
        raise NotImplementedError(INAPPLICABLE_GRID_METHOD_MSG)

    def get_grid_edge_nodes(self, grid: int, edge_nodes: np.ndarray) -> np.ndarray:
        """Get the edge-node connectivity."""
        raise NotImplementedError(INAPPLICABLE_GRID_METHOD_MSG)

    def get_grid_face_edges(self, grid: int, face_edges: np.ndarray) -> np.ndarray:
        """Get the face-edge connectivity."""
        raise NotImplementedError(INAPPLICABLE_GRID_METHOD_MSG)

    def get_grid_face_nodes(self, grid: int, face_nodes: np.ndarray) -> np.ndarray:
        """Get the face-node connectivity."""
        raise NotImplementedError(INAPPLICABLE_GRID_METHOD_MSG)

    def get_grid_nodes_per_face(
        self, grid: int, nodes_per_face: np.ndarray
    ) -> np.ndarray:
        """Get the number of nodes for each face."""
        raise NotImplementedError(INAPPLICABLE_GRID_METHOD_MSG)



"""Module for the soil data IO of PyStemmusScope."""
from collections.abc import Iterable
from pathlib import Path
import hdf5storage
import numpy as np
import xarray as xr
from . import utils
from . import variable_conversion as vc


def _open_multifile_datasets(
    paths: Iterable[Path],
    lat: float,
    lon: float,
    lat_key: str = "lat",
    lon_key: str = "lon",
) -> xr.Dataset:
    """Open mfdatasets, and select location before merging.

    Open multifile netCDF files, and selects the lat & lon
    before merging them by coordinates. xarray's open_mfdataset does not support this
    type of functionality.

    Args:
        paths: Iterable containing the paths to the netCDF files
        lat: Latitude of the site of interest (in degrees North)
        lon: Longitude of the site of interest (in degrees East)
        lat_key: Variable name corresponding to the latitude.
        lon_key: Variable name corresponding to the longitude.

    Returns:
        Dataset containing the merged data for a single location in space.
    """
    datasets: list[xr.Dataset] = []
    for file in paths:
        ds = xr.open_dataset(file)
        #  Drop attributes to avoid combine conflicts
        ds.attrs = ""  # type: ignore
        datasets.append(ds.sel({lat_key: lat, lon_key: lon}, method="nearest"))
    return xr.combine_by_coords(datasets)  # type: ignore


def _read_lambda_coef(
    lambda_directory: Path, lat: float, lon: float, depth_indices: list[int]
) -> dict:
    """Read the lambda coefficient files and return the data in a dict.

    Args:
        lambda_directory: Path to the directory which contains the lambda data.
        lat: Latitude of the site of interest (in degrees North)
        lon: Longitude of the site of interest (in degrees East)
        depth_indices: List of which indices (0 - 7) should be selected from the
            lambda variable dataset.

    Returns:
        dict: Dictionary containing the lambda coefficient data.
    """
    if not np.all([d in range(8) for d in depth_indices]):
        raise ValueError("Incorrect depth indices provided. Indices range from 0 to 7")

    lambda_files = sorted(lambda_directory.glob("lambda_l*.nc"))

    ds = _open_multifile_datasets(lambda_files, lat, lon)

    # which depth indices the STEMMUS_SCOPE model expects
    ds = ds.sortby("depth")  # make sure that the depths are sorted in increasing order
    coef_lambda = ds["lambda"].isel(depth=depth_indices).values

    return {"Coef_Lamda": coef_lambda}


def _read_soil_composition(
    soil_data_path: Path, lat: float, lon: float, depth_indices: list[int]
) -> dict:
    """Read the soil composition files and return them in a dict.

    Args:
        soil_data_path: Path to the directory which contains the soil data.
        lat: Latitude of the site of interest (in degrees North)
        lon: Longitude of the site of interest (in degrees East)
        depth_indices: List of which indices (0 - 7) should be selected from the
            soil composition dataset.

    Returns:
        Dictionary containing the soil composition data.
    """
    soil_comp_fnames = [
        "CLAY1.nc",
        "CLAY2.nc",
        "OC1.nc",
        "OC2.nc",
        "SAND1.nc",
        "SAND2.nc",
        "SILT1.nc",
        "SILT2.nc",
    ]

    soil_comp_paths = [soil_data_path / fname for fname in soil_comp_fnames]

    ds = _open_multifile_datasets(soil_comp_paths, lat, lon)

    if not np.all([d in range(8) for d in depth_indices]):
        raise ValueError("Incorrect depth indices provided. Indices range from 0 to 7")

    ds = ds.sortby("depth")  # make sure that the depths are sorted in increasing order
    ds = ds.isel(depth=depth_indices)

    clay_fraction = ds["CLAY"].values / 100  # convert % to fraction
    sand_fraction = ds["SAND"].values / 100  # convert % to fraction
    organic_fraction = ds["OC"].values / 10000  # convert from 1/100th % to fraction.

    return {"FOC": clay_fraction, "FOS": sand_fraction, "MSOC": organic_fraction}


def _read_hydraulic_parameters(
    soil_data_path: Path, lat: float, lon: float, depths: list[int]
) -> dict:
    """Read the soil hydraulic parameters from the Schaap dataset and return a dict.

    Args:
        soil_data_path: Path to the directory which contains the soil data.
        lat: Latitude of the site of interest (in degrees North)
        lon: Longitude of the site of interest (in degrees East)
        depths: List of depths which should be selected from the dataset. The
            valid depths are: 0, 5, 15, 30, 60, 100 and 200 cm.

    Returns:
        dict: Dictionary containing the hydraulic parameters.
    """
    ptf_files = sorted((soil_data_path / "Schaap").glob("PTF_*.nc"))
    ds = _open_multifile_datasets(
        ptf_files, lat, lon, lat_key="latitude", lon_key="longitude"
    )

    valid_depths = [0, 5, 15, 30, 60, 100, 200]
    if not np.all([d in valid_depths for d in depths]):
        raise ValueError(
            "Incorrect depth value(s) provided. Available depths are" f"{valid_depths}"
        )

    schaap_vars = ["alpha", "Ks", "thetas", "thetar", "n"]
    schaap_data = {key: np.zeros(len(depths)) for key in schaap_vars}

    for i, depth in enumerate(depths):
        for var in schaap_vars:
            schaap_data[var][i] = ds[f"{var}_{depth}cm"]

    fieldmc = vc.field_moisture_content(
        schaap_data["thetar"],
        schaap_data["thetas"],
        schaap_data["alpha"],
        schaap_data["n"],
    )

    return {
        "SaturatedMC": schaap_data["thetas"],
        "ResidualMC": schaap_data["thetar"],
        "Coefficient_n": schaap_data["n"],
        "Coefficient_Alpha": schaap_data["alpha"],
        "porosity": schaap_data["thetas"],
        "Ks0": schaap_data["Ks"][0],
        "SaturatedK": schaap_data["Ks"] / (24 * 3600),  # convert 1/day -> 1/s
        "fieldMC": fieldmc,
        "theta_s0": schaap_data["thetas"][0],
    }


def _read_surface_data(soil_data_path: Path, lat: float, lon: float) -> dict:
    """Read the fmax variable from the surface dataset and return it in a dict.

    Args:
        soil_data_path: Path to the directory which contains the surface data.
        lat: Latitude of the site of interest (in degrees North)
        lon: Longitude of the site of interest (in degrees East)

    Returns:
        dict: Dictionary containing the `fmax` value (maximum fractional saturated area)
    """
    ds = xr.open_dataset(soil_data_path / "surfdata.nc")
    lat, lon = utils.convert_to_lsm_coordinates(lat, lon)
    ds = ds.sel(lsmlat=lat, lsmlon=lon)

    fmax = ds["FMAX"].values

    return {"fmax": fmax}


def _collect_soil_data(soil_data_path: Path, lat: float, lon: float) -> dict:
    """Call and merge all individual data collectors into a single write-ready dict.

    Args:
        soil_data_path: Path to the directory which contains the soil data.
        lat: Latitude of the site of interest (in degrees North)
        lon: Longitude of the site of interest (in degrees East)

    Returns:
        dict: Dictionary containing all the processed soil property data.
    """
    lambda_directory = soil_data_path / "lambda"

    schaap_depths = [0, 5, 30, 60, 100, 200]
    depth_indices = [0, 2, 4, 5, 6, 7]

    matfiledata = _read_lambda_coef(lambda_directory, lat, lon, depth_indices)
    matfiledata.update(
        _read_hydraulic_parameters(soil_data_path, lat, lon, schaap_depths)
    )
    matfiledata.update(_read_soil_composition(soil_data_path, lat, lon, depth_indices))
    matfiledata.update(_read_surface_data(soil_data_path, lat, lon))

    return matfiledata


def _retrieve_latlon(file: Path) -> tuple[float, float]:
    """Retrieve the latitude and longitude coordinates from the dataset file.

    Args:
        file: Full path to the netCDF file containing the site latitude and longitude.

    Returns:
        The latitude and longitude values. Latitude in degrees N,
            longitude in degrees E.
    """
    ds = xr.open_dataset(file)
    lon = ds.longitude.values.flatten()
    lat = ds.latitude.values.flatten()
    return lat, lon


def prepare_soil_data(config: dict) -> None:
    """Prepare the soil input data for the STEMMUS_SCOPE model.

    The data for the input location is parsed, and written to a file that can be easily
    read in by Matlab.

    Args:
        config: The PyStemmusScope configuration dictionary.
    """
    loc, fmt = utils.check_location_fmt(config["Location"])

    if fmt == "site":
        forcing_file = utils.get_forcing_file(config)
        # Data missing at ID-Pag site. See github.com/EcoExtreML/STEMMUS_SCOPE/issues/77
        if config["Location"].startswith("ID"):
            lat, lon = -1.0, 112.0
        else:
            lat, lon = _retrieve_latlon(forcing_file)

    elif fmt == "latlon":
        lat = loc[0]  # type: ignore
        lon = loc[1]  # type: ignore
    else:
        raise NotImplementedError

    matfiledata = _collect_soil_data(Path(config["SoilPropertyPath"]), lat, lon)

    hdf5storage.savemat(
        Path(config["InputPath"]) / "soil_parameters.mat",
        mdict=matfiledata,
        appendmat=False,
    )
    utils.remove_dates_from_header(Path(config["InputPath"]) / "soil_parameters.mat")


def prepare_soil_init(config: dict) -> None:
    """Prepare the soil inital conditions data for the STEMMUS_SCOPE model.

    The data for the input location is parsed, and written to a file that can be easily
    read in by Matlab.

    Args:
        config: The PyStemmusScope configuration dictionary.
    """
    loc, fmt = utils.check_location_fmt(config["Location"])

    if fmt == "site":
        matfiledata = _read_soil_initial_conditions_plumber2(
            soil_init_path=Path(config["InitialConditionPath"]),
            sitename=loc,  # type: ignore
        )
    elif fmt == "latlon":
        matfiledata = _read_soil_initial_conditions_global(
            soil_init_path=Path(config["InitialConditionPath"]),
            lat=loc[0],  # type: ignore
            lon=loc[1],  # type: ignore
            start_time=config["StartTime"],
        )
    else:
        raise NotImplementedError

    hdf5storage.savemat(
        Path(config["InputPath"]) / "soil_init.mat",
        mdict=matfiledata,
        appendmat=False,
    )
    utils.remove_dates_from_header(Path(config["InputPath"]) / "soil_init.mat")


def _extract_soil_initial_variables(soil_init_ds: xr.Dataset):
    """Extract the soil intial variables from the soil init dataset.

    Args:
        soil_init_ds: Dataset containing the following era5-land variables;
            skin_temperature,
            soil_temperature_level_1, soil_temperature_level_2,
            soil_temperature_level_3, soil_temperature_level_4,
            volumetric_soil_water_layer_1, volumetric_soil_water_layer_2,
            volumetric_soil_water_layer_3, volumetric_soil_water_layer_4

    Returns:
        Dictionary containing the STEMMUS_SCOPE variable names (keys) and their intial
            soil condition values.
    """
    return {
        "Tss": float(soil_init_ds["skt"].values) - 273.15,
        "InitT0": float(soil_init_ds["skt"].values) - 273.15,
        "InitT1": float(soil_init_ds["stl1"].values) - 273.15,
        "InitT2": float(soil_init_ds["stl2"].values) - 273.15,
        "InitT3": float(soil_init_ds["stl3"].values) - 273.15,
        "InitT4": float(soil_init_ds["stl4"].values) - 273.15,
        "InitT5": float(soil_init_ds["stl4"].values) - 273.15,
        "InitT6": float(soil_init_ds["stl4"].values) - 273.15,
        "InitX0": float(soil_init_ds["swvl1"].values),
        "InitX1": float(soil_init_ds["swvl1"].values),
        "InitX2": float(soil_init_ds["swvl2"].values),
        "InitX3": float(soil_init_ds["swvl3"].values),
        "InitX4": float(soil_init_ds["swvl4"].values),
        "InitX5": float(soil_init_ds["swvl4"].values),
        "InitX6": float(soil_init_ds["swvl4"].values),
        "BtmX": float(soil_init_ds["swvl4"].values),
    }


def _read_soil_initial_conditions_plumber2(
    soil_init_path: Path,
    sitename: str,
) -> dict[str, float]:
    ds = xr.open_mfdataset(str(soil_init_path / f"{sitename}*.nc"))
    ds = ds.squeeze()  # Remove lat, lon, time dims.
    ds.compute()

    return _extract_soil_initial_variables(ds)


def _read_soil_initial_conditions_global(
    soil_init_path: Path,
    lat: float,
    lon: float,
    start_time: np.datetime64,
) -> dict[str, float]:
    """Read soil initial conditions from era5-land data.

    Args:
        soil_init_path (Path): Path to the global soil init data directory.
        lat (float): Latitude of the location of interest.
        lon (float): Longitude of the location of interest.
        start_time (np.datetime64): Start time of the model.

    Returns:
        Dictionary containing the STEMMUS_SCOPE variable names (keys) and their intial
            soil condition values.
    """
    ds = xr.open_mfdataset(str(soil_init_path / "*.nc"))
    ds = ds.sel(latitude=lat, longitude=lon, method="nearest")
    ds = ds.sel(time=start_time, method="nearest")
    ds.compute()

    return _extract_soil_initial_variables(ds)



"""PyStemmusScope wrapper around Stemmus_Scope."""

import logging
import os
import shlex
import subprocess
from pathlib import Path
from typing import Optional
from typing import Union
from . import config_io
from . import forcing_io
from . import soil_io
from . import utils


logger = logging.getLogger(__name__)


def _is_model_src_exe(model_src_path: Path) -> bool:
    """Check if input exists.

    Returns True if input is a file and False if it is a directory.

    Args:
        model_src_path(Path): path to Stemmus_Scope executable file or to a
        directory containing model source codes.
    """
    if model_src_path.is_file():
        msg = (
            "The model executable file can be used on a Unix system "
            "where MCR is installed, see the "
            "`documentation<https://pystemmusscope.readthedocs.io/>`_."
        )
        logger.info("%s", msg)
        return True
    if model_src_path.is_dir():
        return False
    msg = (
        "Provide a valid path to an executable file or "
        "to a directory containing model source codes, "
        "see the `documentation<https://pystemmusscope.readthedocs.io/>`_."
    )
    raise ValueError(msg)


def _check_interpreter(interpreter: Union[None, str]) -> None:
    if interpreter not in {"Octave", "Matlab"}:
        msg = (
            "Set `interpreter` as Octave or Matlab to run the model using source codes."
            "Otherwise set `model_src_path` to the model executable file, "
            "see the `documentation<https://pystemmusscope.readthedocs.io/>`_."
        )
        raise ValueError(msg)


def _run_sub_process(args: Union[str, list[str]], cwd: Optional[Path] = None) -> str:
    """Run subprocess' Popen, using a list of arguments.

    Args:
        args: Arguments to be run
        cwd: Desired working directory

    Raises:
        subprocess.CalledProcessError: If Popen returns an error code other than 0 or
            139.

    Returns:
        str: Captured stdout.
    """
    result = subprocess.Popen(
        args,
        cwd=cwd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
    )
    exit_code = result.wait()
    stdout, stderr = result.communicate()
    # TODO handle stderr properly
    # when using octave, exit_code might be 139
    # see issue STEMMUS_SCOPE_Processing/issues/46
    if exit_code not in [0, 139]:
        raise subprocess.CalledProcessError(
            returncode=exit_code, cmd=args, stderr=stderr, output=stdout
        )
    if exit_code == 139:
        logger.warning(stderr)

    # TODO return log info line by line!
    logger.info(stdout)
    return stdout.decode("utf-8")


class StemmusScope:
    """PyStemmusScope wrapper around Stemmus_Scope model."""

    def __init__(
        self,
        config_file: Union[str, Path],
        model_src_path: Union[str, Path],
        interpreter: Optional[str] = None,
    ):
        """PyStemmusScope wrapper around Stemmus_Scope model.

        For a detailed model description, look at
        [this publication](https://gmd.copernicus.org/articles/14/1379/2021/).

        Configures the model and prepares forcing and soil data for the model run.

        Arguments:
            config_file: Path to Stemmus_Scope configuration file. An example
                config_file can be found in tests/test_data in [STEMMUS_SCOPE_Processing
                repository](https://github.com/EcoExtreML/STEMMUS_SCOPE_Processing).
            model_src_path: Path to Stemmus_Scope executable file or to a
                directory containing model source codes.
            interpreter (optional): Use `Matlab` or `Octave`. Only required if
                `model_src_path` is a path to model source codes.

        Example:
            See notebooks/run_model_in_notebook.ipynb at the [STEMMUS_SCOPE_Processing
            repository](https://github.com/EcoExtreML/STEMMUS_SCOPE_Processing)
        """
        # make sure paths are abolute and path objects
        config_path = utils.to_absolute_path(config_file)
        model_src = utils.to_absolute_path(model_src_path)

        # check the path to model source
        self.exe_file = None
        if _is_model_src_exe(model_src):
            self.exe_file = model_src
        else:
            _check_interpreter(interpreter)

        self.model_src = model_src
        self.interpreter = interpreter

        # read config template
        self._config = config_io.read_config(config_path)

    def setup(
        self,
        WorkDir: Optional[str] = None,
        Location: Optional[str] = None,
        StartTime: Optional[str] = None,
        EndTime: Optional[str] = None,
    ) -> str:
        """Configure the model run.

        1. Creates config file and input/output directories based on the config template
        2. Prepare forcing and soil data

        Args:
            WorkDir: path to a directory where input/output directories should be
                created.
            Location: Location of the model run. Can be a site ("FI-Hyy") or lat/lon,
                e.g., "(52.0, 4.05)".
            ForcingFileName: forcing file name. Forcing file should be in netcdf format.
            StartTime: Start time of the model run. It must be in
                ISO format (e.g. 2007-01-01T00:00).
            EndTime: End time of the model run. It must be in ISO format
                (e.g. 2007-01-01T00:00).

        Returns:
            Path to the config file
        """
        # update config template if needed
        if WorkDir:
            self._config["WorkDir"] = WorkDir

        if Location:
            self._config["Location"] = Location

        if StartTime:
            self._config["StartTime"] = StartTime

        if EndTime:
            self._config["EndTime"] = EndTime

        # validate config *before* directory creation
        config_io.validate_config(self._config)

        # create customized config file and input/output directories for model run
        _, _, self.cfg_file = config_io.create_io_dir(self._config)

        self._config = config_io.read_config(self.cfg_file)

        forcing_io.prepare_forcing(self._config)
        soil_io.prepare_soil_data(self._config)
        soil_io.prepare_soil_init(self._config)

        return str(self.cfg_file)

    def run(self) -> str:
        """Run model using executable.

        Returns:
            The model log.
        """
        if self.exe_file:
            # run using MCR
            args = [f"{self.exe_file} {self.cfg_file}"]
            # set matlab log dir
            os.environ["MATLAB_LOG_DIR"] = str(self._config["InputPath"])
            result = _run_sub_process(args, None)
        if self.interpreter == "Matlab":
            # set Matlab arguments
            path_to_config = f"'{self.cfg_file}'"
            eval_code = f"STEMMUS_SCOPE_exe({path_to_config});exit;"
            args = ["matlab", "-r", eval_code, "-nodisplay", "-nosplash", "-nodesktop"]

            # seperate args dont work on linux!
            result = _run_sub_process(
                args if utils.os_name() == "nt" else shlex.join(args), self.model_src
            )
        if self.interpreter == "Octave":
            # set Octave arguments
            # use subprocess instead of oct2py,
            # see issue STEMMUS_SCOPE_Processing/issues/46
            path_to_config = f"'{self.cfg_file}'"
            # fix for windows
            path_to_config = path_to_config.replace("\\", "/")
            eval_code = f"STEMMUS_SCOPE_exe({path_to_config});exit;"
            args = ["octave", "--eval", eval_code, "--no-gui", "--silent"]

            # seperate args dont work on linux!
            result = _run_sub_process(
                args if utils.os_name() == "nt" else shlex.join(args), self.model_src
            )
        return result

    @property
    def config(self) -> dict:
        """Return the configurations for this model."""
        return self._config



"""Download land cover data using the cdsapi."""
from pathlib import Path
import cdsapi


with (Path.home() / ".cdsloginrc").open(encoding="utf8") as f:
    uid = f.readline().strip()
    api_key = f.readline().strip()


c = cdsapi.Client(
    url="https://cds.climate.copernicus.eu/api/v2",
    key=f"{uid}:{api_key}",
    verify=True,
)


years = [2013]


for year in years:
    c.retrieve(
        "satellite-land-cover",
        {
            "variable": "all",
            "format": "zip",
            "year": f"{year}",
            "version": "v2.0.7cds",
        },
        f"cds_landcover_{year}.zip",
    )



"""Download monthly ERA5-land data using the cdsapi."""
import itertools
from pathlib import Path
import cdsapi
from pathos.threading import ThreadPool as Pool


N_TRIES = 3


with (Path.home() / ".cdsloginrc").open(encoding="utf8") as f:
    uid = f.readline().strip()
    api_key = f.readline().strip()


def request_lai_data(year, month):
    """Request the LAI data from the CDS."""
    c = cdsapi.Client(
        url="https://cds.climate.copernicus.eu/api/v2",
        key=f"{uid}:{api_key}",
        verify=True,
        retry_max=1,
    )

    c.retrieve(
        "satellite-lai-fapar",
        {
            "format": "zip",
            "variable": "lai",
            "satellite": ("proba" if year >= 2014 else "spot"),
            "sensor": "vgt",
            "horizontal_resolution": "1km",
            "product_version": "V1",
            "year": f"{year}",
            "month": [
                f"{month}",
            ],
            "nominal_day": [
                "10",
                "20",
                "28",
                "30",
                "31",
            ],
        },
        f"LAI_fapar_vgt_{year}-{month}.zip",
    )


if __name__ == "__main__":
    years = list(range(2012, 2018))
    months = [str(x).rjust(2, "0") for x in range(1, 13)]

    _years, _months = map(list, zip(*itertools.product(years, months)))

    pool = Pool(nodes=2)
    pool.map(request_lai_data, _years, _months)



"""PyStemmusScope save module.

Module designed to create a netcdf file following the
[ALMA convention](https://web.lmd.jussieu.fr/~polcher/ALMA/)
from csv files following the
[SCOPE format](https://scope-model.readthedocs.io/en/latest/outfiles.html)
in the output directory.

The file
[`required_netcf_variables.csv`](https://github.com/EcoExtreML/STEMMUS_SCOPE/blob/main/utils/csv_to_nc/required_netcf_variables.csv)
lists required variable names and their attributes based on the
[ALMA+CF convention table](https://docs.google.com/spreadsheets/d/1CA3aTvI9piXqRqO-3MGrsH1vW-Sd87D8iZXHGrqK42o/edit#gid=2085475627).

Note:
    See notebooks/run_model_in_notebook.ipynb in the
    [STEMMUS_SCOPE_Processing repository](https://github.com/EcoExtreML/STEMMUS_SCOPE_Processing).

"""
import logging
from pathlib import Path
import numpy as np
import pandas as pd
import xarray as xr
from PyStemmusScope import config_io
from PyStemmusScope import forcing_io
from PyStemmusScope import utils
from . import variable_conversion as vc


logger = logging.getLogger(__name__)

DATASET_ATTRS = {
    "model": "STEMMUS_SCOPE",
    "institution": "University of Twente; Northwest A&F University",
    "contact": (
        "Zhongbo Su, z.su@utwente.nl; "
        "Yijian Zeng, y.zeng@utwente.nl; "
        "Yunfei Wang, y.wang-3@utwente.nl"
    ),
    "license_type": "CC BY 4.0",
    "license_url": "https://creativecommons.org/licenses/by/4.0/",
}


def _select_forcing_variables(
    forcing_dict: dict, forcing_var: str, alma_var: str
) -> xr.DataArray:
    """Select the variable needed by ALMA convention.

    Args:
        forcing_dict(dict): a dictionary returned by
            `PyStemmusScope.forcing_io.read_forcing_data_plumber2()` or
            `read_forcing_data_global()`
        forcing_var(str): variable name in forcing dataset.
        alma_var(str): variable name in ALMA convention.

    Returns:
        xr.DataArray: a data array which its variable name is alma_name.
    """
    # select the forcing variable
    data_array = forcing_dict[forcing_var]

    # rename the variable name to alma_name
    return data_array.rename(alma_var)


def _prepare_soil_data(csv_file: Path, var_name: str, time: list) -> xr.DataArray:
    """Return simulated soil temperature and soil moisture as `xr.DataArray`.

    Args:
        csv_file: csv file name generated by Stemmus_Scope model.
        var_name: variable name by ALMA convention.
        time: time values to be used for the time coordinates.

    Returns:
        xr.DataArray: a dataarray with two dimensions of time and z.
    """
    # the first two rows are depth and thickness
    data = pd.read_csv(csv_file, delimiter=",", header=[0, 1])

    # skip first row that is unit
    data = data.iloc[1:]

    # make sure it is float and not str
    data = data.astype(float)

    # get depth, thickness info
    depths = []
    thicknesses = []
    for depth, thickness in data.columns:
        depths.append(float(depth))
        thicknesses.append(float(thickness))

    # soil layer metadata
    soil_metadata = _create_soil_layer_metadata(thicknesses, depths)

    # drop thickness
    data = data.droplevel(level=1, axis=1)

    if var_name == "SoilTemp":
        # Celsius to Kelvin : K = 273.15 + C
        data = data + 273.15

    elif var_name == "SoilMoist":
        # cm to m
        thicknesses_m = np.array(thicknesses) / 100.0

        for index in data.index:
            # m3/m3 to kg/m2
            volumetric_water_content = np.array(data.loc[index])
            data.loc[index] = vc.soil_moisture(volumetric_water_content, thicknesses_m)

    # reshape the data frame, it returns Series
    data = data.stack()

    # set values
    layers = range(1, data.index.levshape[1] + 1)
    data.index.names = ["time", "z"]
    data.index = data.index.set_levels([time, layers], level=["time", "z"])
    data.name = var_name

    # convert data to xarray data array
    data_array = data.to_xarray()

    # add z attributes
    data_array["z"].attrs = {
        "long_name": "Soil layer",
        "standard_name": "Soil layer",
        "definition": soil_metadata,
        "units": "-",
    }

    return data_array


def _prepare_simulated_data(
    csv_file: Path, model_name: str, alma_name: str, time: list
) -> xr.DataArray:
    """Return model simulation as `xr.DataArray`.

    Args:
        csv_file: csv file name generated by Stemmus_Scope model.
        model_name: variable name by Stemmus_Scope model.
        alma_name: variable name by ALMA conventions.
        time: time values to be used for the time coordinates.

    Returns:
        xr.DataArray: a dataarray with one dimension of time.
    """
    # the first three rows are names and units
    data = pd.read_csv(csv_file, delimiter=",")

    # select variable and skip first row that is unit
    data = data[model_name].iloc[1:]

    # set time values
    data.index = time
    data.index.names = ["time"]  # type: ignore

    # rename it to alma name
    data.name = alma_name

    # make sure it is float and not str
    data = data.astype("float32")

    # convert dataframe to xarray data array
    return data.to_xarray()


def _create_soil_layer_metadata(
    thicknesses: list[float], depths: list[float]
) -> list[str]:
    """Create soil layer metadata for STEMMUS_SCOPE.

    Note:
        layer_1: 0.0 - 1.0 cm
        layer_2: 1.0 - 2.0 cm
        layer_3: 2.0 - 3.0 cm
    """
    metadata = []
    for index, (thickness, depth) in enumerate(zip(thicknesses, depths)):
        metadata.append(f"layer_{index + 1}: {(depth - thickness)} - {depth} cm")

    return metadata


def _update_dataset_attrs_dims(dataset: xr.Dataset, forcing_dict: dict) -> xr.Dataset:
    """Update dimentions of a dataset according to ALMA conventions.

    Args:
        dataset: Dataset with varaibles in ALMA conventions.
        forcing_dict: a dictionary returned by
            `PyStemmusScope.forcing_io.read_forcing_data_plumber2()` or
            `read_forcing_data_global()`

    Returns:
        The dataset with dimensions ("time", "x", "y").
    """
    # add x/y dims to the dataset
    dataset_expanded = dataset.expand_dims(["x", "y"])

    # change the order of dims
    req_dims = ["time", "x", "y"]
    if any(dim not in dataset_expanded.dims for dim in req_dims):
        raise ValueError("Data should have dimensions time, y, x.")

    if "z" in dataset_expanded.dims:
        dataset_reordered = dataset_expanded.transpose("time", "z", "y", "x")
    else:
        dataset_reordered = dataset_expanded.transpose("time", "y", "x")

    # additional metadata
    lat = forcing_dict["latitude"]
    lon = forcing_dict["longitude"]
    dataset_reordered.attrs = DATASET_ATTRS
    dataset_reordered.attrs["latitude"] = lat
    dataset_reordered.attrs["longitude"] = lon

    # update values of x and y coords
    dataset = dataset_reordered.assign_coords(
        {
            "x": [lon],
            "y": [lat],
        }
    )

    # update x, y attributes
    dataset["x"].attrs = {
        "long_name": "Gridbox longitude",
        "standard_name": "longitude",
        "units": "degrees",
    }

    dataset["y"].attrs = {
        "long_name": "Gridbox latitude",
        "standard_name": "latitude",
        "units": "degrees",
    }

    return dataset


def to_netcdf(config_file: str, cf_filename: str) -> str:
    """Save csv files generated by STEMMUS_SCOPE to a ALMA compliant netCDF file.

    Args:
        config_file: Path to the config file.
        cf_filename: Path to a csv file for ALMA conventions.

    Returns:
        Path to a csv file under the output directory.
    """
    config = config_io.read_config(Path(config_file))
    loc, fmt = utils.check_location_fmt(config["Location"])

    # list of required forcing variables, Alma_short_name: forcing_io_name, # model_name
    var_names = {
        "RH": "rh",  # RH
        "SWdown_ec": "sw_down",  # Rin
        "LWdown_ec": "lw_down",  # Rli
        "Qair": "Qair",
        "Tair": "t_air_celcius",  # Ta
        "Psurf": "psurf_hpa",  # P
        "Wind": "wind_speed",  # u
        "Precip": "precip_conv",  # Pre
    }

    if fmt == "site":
        # read forcing file into a dict
        forcing_dict = forcing_io.read_forcing_data_plumber2(
            utils.get_forcing_file(config),
            config["StartTime"],
            config["EndTime"],
        )
    elif fmt == "latlon":
        forcing_dict = forcing_io.read_forcing_data_global(
            Path(config["ForcingPath"]),
            lat=loc[0],  # type: ignore
            lon=loc[1],  # type: ignore
            start_time=np.datetime64(config["StartTime"]),
            end_time=np.datetime64(config["EndTime"]),
        )

    # get time info
    time = forcing_dict["time"]

    # read convention file
    conventions = pd.read_csv(cf_filename)

    alma_short_names = conventions["short_name_alma"]
    data_list = []
    for alma_name in alma_short_names:
        df = conventions.loc[alma_short_names == alma_name].iloc[0]
        file_name = Path(config["OutputPath"]) / df["file_name_STEMMUS-SCOPE"]

        if alma_name in var_names:
            # select data
            data_array = _select_forcing_variables(
                forcing_dict, var_names[alma_name], alma_name
            )

        # create data array
        elif alma_name in {"SoilTemp", "SoilMoist"}:
            data_array = _prepare_soil_data(file_name, alma_name, time)
        else:
            data_array = _prepare_simulated_data(
                file_name, df["short_name_STEMMUS-SCOPE"], alma_name, time
            )

        # update attributes of array
        data_array.attrs = {
            "units": df["unit"],
            "long_name": df["long_name"],
            "standard_name": df["standard_name"],
            "STEMMUS-SCOPE_name": df["short_name_STEMMUS-SCOPE"],
            "definition": df["definition"],
        }

        # add to list
        data_list.append(data_array)

    # merge to a dataset
    dataset = xr.merge(data_list)

    # update dimensions
    dataset = _update_dataset_attrs_dims(dataset, forcing_dict)

    # for writing to netcdf, time attrs should be added
    # time attrs should be the same as plumber 2 forcing data
    # otherwise it cannot be uploaded to modelevaluation portal
    start_time = time.dt.strftime("%Y-%m-%d").values[0]
    time_encode = {
        "time": {"units": f"seconds since {start_time}", "calendar": "standard"}
    }
    # save to nc file
    nc_filename = (
        Path(config["OutputPath"])
        / f"{Path(config['OutputPath']).stem}_STEMMUS_SCOPE.nc"
    )
    dataset.to_netcdf(path=nc_filename, encoding=time_encode)

    return str(nc_filename)



"""Utility functions for making the docker process work."""
import warnings
from pathlib import Path
from PyStemmusScope.config_io import read_config


try:
    import docker
except ImportError:
    docker = None


def make_docker_vols_binds(cfg_file: str) -> tuple[list[str], list[str]]:
    """Make docker volume mounting configs.

    Args:
        cfg_file: Location of the config file

    Returns:
        volumes, binds
    """
    cfg = read_config(cfg_file)
    cfg_dir = Path(cfg_file).parent
    volumes = []
    binds = []

    # Make sure no subpaths are mounted:
    if not cfg_dir.is_relative_to(cfg["InputPath"]):
        volumes.append(str(cfg_dir))
        binds.append(f"{str(cfg_dir)}:{str(cfg_dir)}")
    if (not Path(cfg["InputPath"]).is_relative_to(cfg_dir)) or (
        Path(cfg["InputPath"]) == cfg_dir
    ):
        volumes.append(cfg["InputPath"])
        binds.append(f"{cfg['InputPath']}:{cfg['InputPath']}")
    if not Path(cfg["OutputPath"]).is_relative_to(cfg_dir):
        volumes.append(cfg["OutputPath"])
        binds.append(f"{cfg['OutputPath']}:{cfg['OutputPath']}")

    return volumes, binds


def check_tags(image: str, compatible_tags: tuple[str, ...]):
    """Check if the tag is compatible with this version of the BMI.

    Args:
        image: The full image name (including tag)
        compatible_tags: Tags which are known to be compatible with this version of the
            BMI.
    """
    if ":" not in image:
        msg = (
            "Could not validate the Docker image tag, as no tag was provided.\n"
            "Please set the Docker image tag in the configuration file."
        )
        warnings.warn(UserWarning(msg), stacklevel=1)

    tag = image.split(":")[-1]
    if tag not in compatible_tags:
        msg = (
            f"Docker image tag '{tag}' not found in compatible tags "
            f"({compatible_tags}).\n"
            "You might experience issues or unexpected results."
        )
        warnings.warn(UserWarning(msg), stacklevel=1)


def find_image(image: str) -> None:
    """See if the desired image is available, and if not, try to pull it."""
    client = docker.APIClient()
    images = client.images()
    tags = []
    for img in images:
        for tag in img["RepoTags"]:
            tags.append(tag)
    if image not in set(tags):
        pull_image(image)


def pull_image(image: str) -> None:
    """Pull the image from ghcr/dockerhub."""
    if ":" in image:
        image, tag = image.split(":")
    else:
        tag = None
    client = docker.from_env()
    image = client.images.pull(image, tag)



"""Documentation about PyStemmusScope."""
import logging
from .stemmus_scope import StemmusScope


logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = ["StemmusScope"]

__author__ = "Sarah Alidoost"
__email__ = "f.alidoost@esciencecenter.nl"
__version__ = "0.4.0"



"""The Docker STEMMUS_SCOPE model process wrapper."""
import os
import socket as pysocket
import warnings
from time import sleep
from typing import Any
from PyStemmusScope.bmi.docker_utils import check_tags
from PyStemmusScope.bmi.docker_utils import find_image
from PyStemmusScope.bmi.docker_utils import make_docker_vols_binds
from PyStemmusScope.bmi.utils import MATLAB_ERROR
from PyStemmusScope.bmi.utils import PROCESS_FINALIZED
from PyStemmusScope.bmi.utils import PROCESS_READY
from PyStemmusScope.bmi.utils import MatlabError
from PyStemmusScope.config_io import read_config


try:
    import docker
except ImportError:
    docker = None


def _model_is_ready(socket: Any, client: Any, container_id: Any) -> None:
    return _wait_for_model(PROCESS_READY, socket, client, container_id)


def _model_is_finalized(socket: Any, client: Any, container_id: Any) -> None:
    return _wait_for_model(PROCESS_FINALIZED, socket, client, container_id)


def _wait_for_model(phrase: bytes, socket: Any, client: Any, container_id: Any) -> None:
    """Wait for the model to be ready to receive (more) commands, or is finalized."""
    output = b""

    while phrase not in output:
        try:
            data = socket.read(1)
        except TimeoutError as err:
            client.stop(container_id)
            logs = client.logs(container_id).decode("utf-8")
            msg = (
                f"Container connection timed out '{container_id['Id']}'."
                f"\nPlease inspect logs:\n{logs}"
            )
            raise TimeoutError(msg) from err

        if data is None:
            msg = "Could not read data from socket. Docker container might be dead."
            raise ConnectionError(msg)
        else:
            output += bytes(data)

        if MATLAB_ERROR in output:
            client.stop(container_id)
            logs = client.logs(container_id).decode("utf-8")
            msg = (
                f"Error in container '{container_id['Id']}'.\n"
                f"Please inspect logs:\n{logs}"
            )
            raise MatlabError(msg)


def _attach_socket(client, container_id) -> Any:
    """Attach a socket to a container and add a timeout to it."""
    connection_timeout = 30  # seconds

    socket = client.attach_socket(container_id, {"stdin": 1, "stdout": 1, "stream": 1})
    if isinstance(socket, pysocket.SocketIO):
        socket._sock.settimeout(connection_timeout)  # type: ignore
    else:
        warnings.warn(
            message=(
                "Unknown socket type found. This might cause issues with the Docker"
                " connection. \nPlease report this to the developers in an issue "
                "on: https://github.com/EcoExtreML/STEMMUS_SCOPE_Processing/issues"
            ),
            stacklevel=1,
        )
    return socket


class StemmusScopeDocker:
    """Communicate with a STEMMUS_SCOPE Docker container."""

    # Default image, can be overridden with config:
    compatible_tags = ("1.5.0",)

    _process_ready_phrase = b"Select BMI mode:"
    _process_finalized_phrase = b"Finished clean up."

    def __init__(self, cfg_file: str):
        """Create the Docker container.."""
        self.cfg_file = cfg_file
        config = read_config(cfg_file)

        self.image = config["DockerImage"]
        find_image(self.image)
        check_tags(self.image, self.compatible_tags)

        self.client = docker.APIClient()

        vols, binds = make_docker_vols_binds(cfg_file)
        self.container_id = self.client.create_container(
            self.image,
            stdin_open=True,
            tty=True,
            detach=True,
            user=f"{os.getuid()}:{os.getgid()}",  # ensure correct user for writing files.
            volumes=vols,
            host_config=self.client.create_host_config(binds=binds),
        )

        self.running = False

    def _wait_for_model(self) -> None:
        """Wait for the model to be ready to receive (more) commands."""
        _model_is_ready(self.socket, self.client, self.container_id)

    def is_alive(self) -> bool:
        """Return if the process is alive."""
        return self.running

    def initialize(self) -> None:
        """Initialize the model and wait for it to be ready."""
        if self.is_alive():
            self.client.stop(self.container_id)

        self.client.start(self.container_id)
        self.socket = _attach_socket(self.client, self.container_id)

        self._wait_for_model()
        os.write(
            self.socket.fileno(),
            bytes(f'initialize "{self.cfg_file}"\n', encoding="utf-8"),
        )
        self._wait_for_model()

        self.running = True

    def update(self) -> None:
        """Update the model and wait for it to be ready."""
        if self.is_alive():
            os.write(self.socket.fileno(), b"update\n")
            self._wait_for_model()
        else:
            msg = "Docker container is not alive. Please restart the model."
            raise ConnectionError(msg)

    def finalize(self) -> None:
        """Finalize the model."""
        if self.is_alive():
            os.write(self.socket.fileno(), b"finalize\n")
            _model_is_finalized(
                self.socket,
                self.client,
                self.container_id,
            )
            sleep(0.5)  # Ensure the container can stop cleanly.
            self.client.stop(self.container_id)
            self.running = False
            self.client.remove_container(self.container_id, v=True)
        else:
            pass
